import os
import os.path as op
import pprint
import shutil
import sys
from difflib import unified_diff
from glob import glob as _sys_glob
from pathlib import Path
from textwrap import dedent

import asdf
import astropy.table
import astropy.time
import deepdiff
import gwcs
import numpy as np
import requests
from astropy.units import Quantity
from ci_watson.artifactory_helpers import (
    BigdataError,
    check_url,
    get_bigdata,
    get_bigdata_root,
)
from deepdiff.operator import BaseOperator
from gwcs.wcstools import grid_from_bounding_box

# from romancal.lib.suffix import replace_suffix
from romancal.stpipe import RomanStep

# Define location of default Artifactory API key, for Jenkins use only
ARTIFACTORY_API_KEY_FILE = "/eng/ssb2/keys/svc_rodata.key"

# Define a request timeout in seconds
TIMEOUT = 30


class RegtestData:
    """Defines data paths on Artifactory and data retrieval methods"""

    def __init__(
        self,
        env="dev",
        inputs_root="roman-pipeline",
        results_root="roman-pipeline-results",
        docopy=True,
        input=None,
        input_remote=None,
        output=None,
        truth=None,
        truth_remote=None,
        remote_results_path=None,
        test_name=None,
        traceback=None,
        **kwargs,
    ):
        self._env = env
        self._inputs_root = inputs_root
        self._results_root = results_root
        self._bigdata_root = get_bigdata_root()

        self.docopy = docopy

        # Initialize @property attributes
        self.input = input
        self.input_remote = input_remote
        self.output = output
        self.truth = truth
        self.truth_remote = truth_remote

        # No @properties for the following attributes
        self.remote_results_path = remote_results_path
        self.test_name = test_name
        self.traceback = traceback

        # Initialize non-initialized attributes
        self.asn = None

    def __repr__(self):
        return pprint.pformat(
            dict(
                input=self.input,
                output=self.output,
                truth=self.truth,
                input_remote=self.input_remote,
                truth_remote=self.truth_remote,
                remote_results_path=self.remote_results_path,
                test_name=self.test_name,
                traceback=self.traceback,
            ),
            indent=1,
        )

    @property
    def input_remote(self):
        if self._input_remote is not None:
            return os.path.join(*self._input_remote)
        else:
            return None

    @input_remote.setter
    def input_remote(self, value):
        if value:
            self._input_remote = value.split(os.sep)
        else:
            self._input_remote = value

    @property
    def truth_remote(self):
        if self._truth_remote is not None:
            return os.path.join(*self._truth_remote)
        else:
            return None

    @truth_remote.setter
    def truth_remote(self, value):
        if value:
            self._truth_remote = value.split(os.sep)
        else:
            self._truth_remote = value

    @property
    def input(self):
        return self._input

    @input.setter
    def input(self, value):
        if value:
            self._input = os.path.abspath(value)
        else:
            self._input = value

    @property
    def truth(self):
        return self._truth

    @truth.setter
    def truth(self, value):
        if value:
            self._truth = os.path.abspath(value)
        else:
            self._truth = value

    @property
    def output(self):
        return self._output

    @output.setter
    def output(self, value):
        if value:
            self._output = os.path.abspath(value)
        else:
            self._output = value

    @property
    def bigdata_root(self):
        return self._bigdata_root

    # The methods
    def get_data(self, path=None, docopy=None):
        """Copy data from Artifactory remote resource to the CWD

        Updates self.input and self.input_remote upon completion
        """
        if path is None:
            path = self.input_remote
        else:
            self.input_remote = path
        if docopy is None:
            docopy = self.docopy
        self.input = get_bigdata(self._inputs_root, self._env, path, docopy=docopy)
        self.input_remote = os.path.join(self._inputs_root, self._env, path)

        return self.input

    def data_glob(self, path=None, glob="*", docopy=None):
        """Get a list of files"""
        if path is None:
            path = self.input_remote
        else:
            self.input_remote = path
        if docopy is None:
            docopy = self.docopy

        # Get full path and proceed depending on whether
        # is a local path or URL.
        root = self.bigdata_root
        if op.exists(root):
            root_path = op.join(root, self._inputs_root, self._env)
            root_len = len(root_path) + 1
            path = op.join(root_path, path)
            file_paths = _data_glob_local(path, glob)
        elif check_url(root):
            root_len = len(self._env) + 1
            file_paths = _data_glob_url(
                self._inputs_root, self._env, path, glob, root=root
            )
        else:
            raise BigdataError(f"Path cannot be found: {path}")

        # Remove the root from the paths
        file_paths = [file_path[root_len:] for file_path in file_paths]
        return file_paths

    def get_truth(self, path=None, docopy=None):
        """Copy truth data from Artifactory remote resource to the CWD/truth

        Updates self.truth and self.truth_remote on completion
        """
        if path is None:
            path = self.truth_remote
        else:
            self.truth_remote = path
        if docopy is None:
            docopy = self.docopy
        os.makedirs("truth", exist_ok=True)
        os.chdir("truth")
        try:
            self.truth = get_bigdata(self._inputs_root, self._env, path, docopy=docopy)
            self.truth_remote = os.path.join(self._inputs_root, self._env, path)
        except BigdataError:
            os.chdir("..")
            raise
        os.chdir("..")

        return self.truth

    # def get_asn(self, path=None, docopy=True, get_members=True):
    #     """Copy association and association members from Artifactory remote
    #     resource to the CWD/truth.
    #
    #     Updates self.input and self.input_remote upon completion
    #
    #     Parameters
    #     ----------
    #     path: str
    #         The remote path
    #
    #     docopy : bool
    #         Switch to control whether or not to copy a file
    #         into the test output directory when running the test.
    #         If you wish to open the file directly from remote
    #         location or just to set path to source, set this to `False`.
    #         Default: `True`
    #
    #     get_members: bool
    #         If an association is the input, retrieve the members.
    #         Otherwise, do not.
    #     """
    #     if path is None:
    #         path = self.input_remote
    #     else:
    #         self.input_remote = path
    #     if docopy is None:
    #         docopy = self.docopy
    #
    #     # Get the association JSON file
    #     self.input = get_bigdata(self._inputs_root, self._env, path,
    #         docopy=docopy)
    #     with open(self.input) as fp:
    #         asn = load_asn(fp)
    #         self.asn = asn
    #
    #     # Get each member in the association as well
    #     if get_members:
    #         for product in asn['products']:
    #             for member in product['members']:
    #                 fullpath = os.path.join(
    #                     os.path.dirname(self.input_remote),
    #                     member['expname'])
    #                 get_bigdata(self._inputs_root, self._env, fullpath,
    #                             docopy=self.docopy)

    def to_asdf(self, path):
        tree = eval(str(self))
        af = asdf.AsdfFile(tree=tree)
        af.write_to(path)

    @classmethod
    def open(cls, filename):
        with asdf.open(filename) as af:
            return cls(**af.tree)


def run_step_from_dict(rtdata, **step_params):
    """Run Steps with given parameter

    Parameters
    ----------
    rtdata: RegtestData
        The artifactory instance

    step_params: dict
        The parameters defining what step to run with what input

    Returns
    -------
    rtdata: RegtestData
        Updated `RegtestData` object with inputs set.

    Notes
    -----
    `step_params` looks like this:
    {
        'input_path': str or None  # The input file path, relative to artifactory
        'step': str                # The step to run, either a class or a config file
        'args': list,              # The arguments passed to `Step.from_cmdline`
    }
    """

    # Get the data. If `step_params['input_path]` is not
    # specified, the presumption is that `rtdata.input` has
    # already been retrieved.
    # input_path = step_params.get('input_path', None)
    # if input_path:
    #     try:
    #         rtdata.get_asn(input_path)
    #     except AssociationNotValidError:
    #         rtdata.get_data(input_path)

    # Figure out whether we have a config or class
    step = step_params["step"]
    if step.endswith((".asdf", ".cfg")):
        step = os.path.join("config", step)

    # Run the step
    full_args = [step, rtdata.input]
    full_args.extend(step_params["args"])

    RomanStep.from_cmdline(full_args)

    return rtdata


def run_step_from_dict_mock(rtdata, source, **step_params):
    """Pretend to run Steps with given parameter but just copy data

    For long running steps where the result already exists, just
    copy the data from source

    Parameters
    ----------
    rtdata: RegtestData
        The artifactory instance

    step_params: dict
        The parameters defining what step to run with what input

    source: Path-like folder
        The folder to copy from. All regular files are copied.

    Returns
    -------
    rtdata: RegtestData
        Updated `RegtestData` object with inputs set.

    Notes
    -----
    `step_params` looks like this:
    {
        'input_path': str or None  # The input file path, relative to artifactory
        'step': str                # The step to run, either a class or a config file
        'args': list,              # The arguments passed to `Step.from_cmdline`
    }
    """

    # Get the data. If `step_params['input_path]` is not
    # specified, the presumption is that `rtdata.input` has
    # already been retrieved.
    # input_path = step_params.get('input_path', None)
    # if input_path:
    #     try:
    #         rtdata.get_asn(input_path)
    #     except AssociationNotValidError:
    #         rtdata.get_data(input_path)

    # Copy the data
    for file_name in os.listdir(source):
        file_path = os.path.join(source, file_name)
        if os.path.isfile(file_path):
            shutil.copy(file_path, ".")

    return rtdata


def is_like_truth(rtdata, ignore_asdf_paths, output, truth_path, is_suffix=True):
    """Compare step outputs with truth

    Parameters
    ----------
    rtdata: RegtestData
        The artifactory object from the step run.

    ignore_asdf_paths: dict
        The asdf `diff` keyword arguments

    output: str
        The suffix or full file name to check on.

    truth_path: str
        Location of the truth files.

    is_suffix: bool
        Interpret `output` as just a suffix on the expected output root.
        Otherwise, assume it is a full file name
    """
    __tracebackhide__ = True
    # If given only a suffix, get the root to change the suffix of.
    # If the input was an association, the output should be the name of
    # the product. Otherwise, output is based on input.
    if is_suffix:
        # suffix = output
        if rtdata.asn:
            output = rtdata.asn["products"][0]["name"]
        else:
            output = os.path.splitext(os.path.basename(rtdata.input))[0]
        # output = replace_suffix(output, suffix) + '.asdf'
    rtdata.output = output

    rtdata.get_truth(os.path.join(truth_path, output))

    # diff = FITSDiff(rtdata.output, rtdata.truth, **fitsdiff_default_kwargs)
    report = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    assert report is None, report


def text_diff(from_path, to_path):
    """Assertion helper for diffing two text files

    Parameters
    ----------
    from_path: str
        File to diff from.

    to_path: str
        File to diff to.  The truth.

    Returns
    -------
    diffs: [str[,...]]
        A generator of a list of strings that are the differences.
        The output from `difflib.unified_diff`
    """
    __tracebackhide__ = True
    with open(from_path) as fh:
        from_lines = fh.readlines()
    with open(to_path) as fh:
        to_lines = fh.readlines()

    diffs = unified_diff(from_lines, to_lines, from_path, to_path)

    diff = list(diffs)
    if len(diff) > 0:
        diff.insert(0, "\n")
        raise AssertionError("".join(diff))
    else:
        return True


def _data_glob_local(*glob_parts):
    """Perform a glob on the local path

    Parameters
    ----------
    glob_parts: (path-like,[...])
        List of components that will be built into a single path

    Returns
    -------
    file_paths: [str[, ...]]
        Full file paths that match the glob criterion
    """
    full_glob = Path().joinpath(*glob_parts)
    return _sys_glob(str(full_glob))


def _data_glob_url(*url_parts, root=None):
    """
    Parameters
    ----------
    url: (str[,...])
        List of components that will be used to create a URL path

    root: str
        The root server path to the Artifactory server.
        Normally retrieved from `get_bigdata_root`.

    Returns
    -------
    url_paths: [str[, ...]]
        Full URLS that match the glob criterion
    """
    # Fix root root-ed-ness
    if root.endswith("/"):
        root = root[:-1]

    # Access
    try:
        envkey = os.environ["API_KEY_FILE"]
    except KeyError:
        envkey = ARTIFACTORY_API_KEY_FILE

    try:
        with open(envkey) as fp:
            headers = {"X-JFrog-Art-Api": fp.readline().strip()}
    except (PermissionError, FileNotFoundError):
        print(
            "Warning: Anonymous Artifactory search requests are limited to "
            "1000 results. Use an API key and define API_KEY_FILE environment"
            "variable to get full search results.",
            file=sys.stderr,
        )
        headers = None

    search_url = "/".join([root, "api/search/pattern"])

    # Join and re-split the url so that every component is identified.
    url = "/".join([root] + [idx for idx in url_parts])
    all_parts = url.split("/")

    # Pick out "roman-pipeline", the repo name
    repo = all_parts[4]

    # Format the pattern
    pattern = repo + ":" + "/".join(all_parts[5:])

    # Make the query
    params = {"pattern": pattern}
    with requests.get(search_url, params=params, headers=headers, timeout=TIMEOUT) as r:
        url_paths = r.json()["files"]

    return url_paths


class NDArrayTypeOperator(BaseOperator):
    def __init__(self, rtol=1e-05, atol=1e-08, equal_nan=True, **kwargs):
        super().__init__(**kwargs)
        self.rtol = rtol
        self.atol = atol
        self.equal_nan = equal_nan

    def _compare_arrays(self, a, b):
        difference = {}
        if a.shape != b.shape:
            difference["shapes"] = [a.shape, b.shape]
        if a.dtype != b.dtype:
            difference["dtypes"] = [a.dtype, b.dtype]
        if isinstance(a, Quantity) and isinstance(b, Quantity):
            if a.unit != b.unit:
                difference["units"] = [a.unit, b.unit]
        if a.dtype.names or b.dtype.names:
            if a.dtype.names != b.dtype.names:
                difference["names"] = [a.dtype.names, b.dtype.names]
            if not difference:
                # only compare values if no other difference exists
                column_differences = {}
                for name in a.dtype.names:
                    column_difference = self._compare_arrays(a[name], b[name])
                    if column_difference:
                        column_differences[name] = column_difference
                if column_differences:
                    difference["column_values"] = column_differences
            return difference
        if not difference:  # only compare if shapes and dtypes match
            if not np.allclose(
                a, b, rtol=self.rtol, atol=self.atol, equal_nan=self.equal_nan
            ):
                abs_diff = np.abs(a - b)
                index = np.unravel_index(np.nanargmax(abs_diff), a.shape)
                difference["worst_abs_diff"] = {
                    "index": index,
                    "value": abs_diff[index],
                }
                with np.errstate(invalid="ignore", divide="ignore"):
                    # 0 / 0 == nan and produces an 'invalid' error
                    # 1 / 0 == inf and produces a 'divide' error
                    # ignore these here for computing the fractional diff
                    fractional_diff = np.abs(a / b)
                index = np.unravel_index(np.nanargmax(fractional_diff), a.shape)
                difference["worst_fractional_diff"] = {
                    "index": index,
                    "value": fractional_diff[index],
                }
                difference["abs_diff"] = np.nansum(np.abs(a - b))
                difference["n_diffs"] = np.count_nonzero(
                    np.logical_not(
                        np.isclose(
                            a,
                            b,
                            rtol=self.rtol,
                            atol=self.atol,
                            equal_nan=self.equal_nan,
                        )
                    )
                )
        return difference

    def give_up_diffing(self, level, diff_instance):
        difference = self._compare_arrays(level.t1, level.t2)
        if difference:
            diff_instance.custom_report_result("arrays_differ", level, difference)
        return True


class TableOperator(NDArrayTypeOperator):
    def give_up_diffing(self, level, diff_instance):
        a = level.t1.as_array()
        b = level.t2.as_array()
        difference = self._compare_arrays(a, b)
        # also compare meta  for tables
        if level.t1.meta != level.t2.meta:
            meta_difference = deepdiff.DeepDiff(
                level.t1.meta,
                level.t2.meta,
                ignore_nan_inequality=self.equal_nan,
                math_epsilon=self.atol,
                exclude_paths=[
                    "root['date']",
                    "root['version']",
                ],
                # creation date sometimes stored here
            )
            if meta_difference:
                difference["metas_differ"] = meta_difference
        if difference:
            diff_instance.custom_report_result("tables_differ", level, difference)
        return True


class TimeOperator(BaseOperator):
    def give_up_diffing(self, level, diff_instance):
        if level.t1 != level.t2:
            diff_instance.custom_report_result(
                "times_differ",
                level,
                {
                    "difference": level.t1 - level.t2,
                },
            )
        return True


def _wcs_to_ra_dec(wcs):
    x, y = grid_from_bounding_box(wcs.bounding_box)
    return wcs(x, y)


class WCSOperator(NDArrayTypeOperator):
    def give_up_diffing(self, level, diff_instance):
        # for comparing wcs instances this function evaluates
        # each wcs and compares the resulting ra and dec outputs
        # TODO should we compare the bounding boxes?
        ra_a, dec_a = _wcs_to_ra_dec(level.t1)
        ra_b, dec_b = _wcs_to_ra_dec(level.t2)
        ra_diff = self._compare_arrays(ra_a, ra_b)
        dec_diff = self._compare_arrays(dec_a, dec_b)
        difference = {}
        if ra_diff:
            difference["ra"] = ra_diff
        if dec_diff:
            difference["dec"] = dec_diff
        if difference:
            diff_instance.custom_report_result("wcs_differ", level, difference)
        return True


class DiffResult:
    def __init__(self, diff, result, truth):
        self.diff = diff
        self.result = result
        self.truth = truth

    @property
    def identical(self):
        return not self.diff

    @staticmethod
    def _model_type(file):
        import roman_datamodels as rdm

        try:
            with rdm.open(file) as mdl:
                return mdl.__class__.__name__
        except Exception:
            return "Not a model"

    def report(self, **kwargs):
        title = dedent(
            f"""
            Diff report for:
                result file: {self.result}
                    model type: {self._model_type(self.result)}
                truth file: {self.truth}
                    model type: {self._model_type(self.result)}
            """
        )
        report = pprint.pformat(self.diff)

        return f"{title}\n{report}"


def compare_asdf(result, truth, ignore=None, rtol=1e-05, atol=1e-08, equal_nan=True):
    """
    Compare 2 asdf files: result and truth. Note that this comparison is
    asymmetric (swapping result and truth will give a different result).

    Parameters
    ----------

    result : str
        Filename of result asdf file

    truth : str
        Filename of truth asdf file

    ignore : list
        List of tree node paths to ignore during the comparison

    rtol : float
        rtol argument passed to `numpyp.allclose`

    atol : float
        atol argument passed to `numpyp.allclose` and ``DeepDiff``
        as ``math_epsilon``.

    equal_nan : bool
        Ignore nan inequality

    Returns
    -------

    diff_result : DiffResult
        result of the comparison
    """
    exclude_paths = []
    ignore = ignore or []
    for path in ignore:
        key_path = "".join([f"['{k}']" for k in path.split(".")])
        exclude_paths.append(f"root{key_path}")
    operators = [
        NDArrayTypeOperator(
            rtol,
            atol,
            equal_nan,
            types=[asdf.tags.core.NDArrayType, np.ndarray],
        ),
        TimeOperator(types=[astropy.time.Time]),
        TableOperator(rtol, atol, equal_nan, types=[astropy.table.Table]),
        WCSOperator(rtol, atol, equal_nan, types=[gwcs.WCS]),
    ]
    # warnings can be seen in regtest runs which indicate
    # that ddtrace logs are evaluated at times after the below
    # with statement exits resulting in access attempts on the
    # closed asdf file. To try and avoid that we disable
    # lazy loading and memmory mapping
    open_kwargs = {
        "lazy_load": False,
        "copy_arrays": True,
    }
    with (
        asdf.open(result, **open_kwargs) as af0,
        asdf.open(truth, **open_kwargs) as af1,
    ):
        # swap the inputs here so DeepDiff(truth, result)
        # this will create output with 'new_value' referring to
        # the value in the result and 'old_value' referring to the truth
        diff = deepdiff.DeepDiff(
            af1.tree,
            af0.tree,
            ignore_nan_inequality=equal_nan,
            custom_operators=operators,
            exclude_paths=exclude_paths,
            math_epsilon=atol,
            ignore_type_in_groups=[asdf.tags.core.NDArrayType, np.ndarray],
        )
        return DiffResult(diff, result, truth)



""""Diff and compare associations"""

import logging
import re
from collections import Counter, UserList
from copy import copy

from romancal.associations.load_asn import load_asn

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# #########################
# Define the types of diffs
# #########################
class DiffError(AssertionError):
    """Base Class for difference errors"""


class CandidateLevelError(DiffError):
    """Candidate level mismatch"""


class DifferentProductSetsError(DiffError):
    """Different product sets between groups of associations"""


class DuplicateProductError(DiffError):
    """Duplicate products found"""


class MemberMismatchError(DiffError):
    """Membership does not match"""


class TypeMismatchError(DiffError):
    """Association type mismatch"""


class MultiDiffError(UserList, DiffError):
    """List of diff errors"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __str__(self):
        message = ["Following diffs found:\n"]
        for diff in self:
            message.extend(["\n****\n", str(diff), "\n"])
        return "".join(message)


def compare_asn_files(left_paths, right_paths):
    """Compare association files

    Parameters
    ----------
    left_paths: [Path or str[, ...]]
        Set of association files

    right_paths: [Path or str[, ...]]
        Set of association files to compare against

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    __tracebackhide__ = True
    # Read in all the associations, separating out the products into separate
    # associations.
    left_asns = []
    for path in left_paths:
        with open(path) as fh:
            asn = load_asn(fh)
        single_product_asns = separate_products(asn)

        # Collect the associations.
        left_asns += single_product_asns

    right_asns = []
    for path in right_paths:
        with open(path) as fh:
            asn = load_asn(fh)
        single_product_asns = separate_products(asn)

        # Collect the associations.
        right_asns += single_product_asns

    # Compare the associations.
    compare_asn_lists(left_asns, right_asns)


def compare_asn_lists(left_asns, right_asns):
    """Compare to lists of associations

    Both association lists must contain associations that only have
    single products. Use `separate_products` prior to calling this
    function.

    Parameters
    ----------
    left_asns: [`Association`[, ...]]
        Group of associations

    right_asns: [`Association`[, ...]]
        Group of associations to compare

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    __tracebackhide__ = True
    diffs = MultiDiffError()

    # Ensure that product names are unique
    left_product_names, left_duplicates = get_product_names(left_asns)
    right_product_names, right_duplicates = get_product_names(right_asns)
    if left_duplicates:
        diffs.append(
            DuplicateProductError(
                f"Left associations have duplicate products {left_duplicates}"
            )
        )
    if right_duplicates:
        diffs.append(
            DuplicateProductError(
                f"Right associations have duplicate products {right_duplicates}"
            )
        )

    # Ensure that the product name lists are the same.
    name_diff = left_product_names ^ right_product_names
    if name_diff:
        diffs.append(
            DifferentProductSetsError(
                "Left and right associations do not share a common set of products:"
                f" {name_diff}"
            )
        )

    # Compare like product associations
    left_asns_by_product = {asn["products"][0]["name"]: asn for asn in left_asns}
    right_asns_by_product = {asn["products"][0]["name"]: asn for asn in right_asns}
    for product_name in left_product_names:
        try:
            compare_asns(
                left_asns_by_product[product_name], right_asns_by_product[product_name]
            )
        except MultiDiffError as compare_diffs:
            diffs.extend(compare_diffs)
        except KeyError:
            # Most likely due to a previous error. Ignore
            pass

    if diffs:
        raise diffs


def compare_asns(left, right):
    """Compare two associations

    This comparison will include metadata such as
    `asn_type` and membership

    Parameters
    ---------
    left, right : dict
        Two, individual, associations to compare

    Raises
    ------
    MultiDiffError
        If there is a difference.
    """
    _compare_asns(left, right)


def _compare_asns(left, right):
    """Compare two associations

    This comparison will include metadata such as
    `asn_type` and membership

    Parameters
    ---------
    left, right : dict
        Two, individual, associations to compare

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.

    Note
    ----
    This comparison is dependent on the associations being JWST-like associations.
    The attributes that are compared are as follows:

        - key `asn_type`
        - key `products`. Specifically the following are compared:
            - Length of the list
            - key `name` for each product
            - key `members` for each product
        - For the member lists of each product, the following are compared:
            - Length of the list
            - key `expname` for each member
            - key 'exptype` for each member
    """
    diffs = MultiDiffError()

    # Assert that the same result type is the same.
    if left["asn_type"] != right["asn_type"]:
        diffs.append(
            TypeMismatchError(
                f"Type mismatch {left['asn_type']} != {right['asn_type']}"
            )
        )

    # Assert that the level of association candidate is the same.
    # Cannot guarantee value, but that the 'a'/'c'/'o' levels are similar.
    if left["asn_id"][0] != right["asn_id"][0]:
        diffs.append(
            CandidateLevelError(
                f"Candidate level mismatch left '{left['asn_id'][0]}' != right"
                f" '{right['asn_id'][0]}'"
            )
        )

    # Membership
    try:
        compare_membership(left, right)
    except MultiDiffError as compare_diffs:
        diffs.extend(compare_diffs)

    if diffs:
        raise diffs


def compare_membership(left, right):
    """Compare two associations' membership

    Parameters
    ---------
    left, right : dict
        Two, individual, associations to compare

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    diffs = MultiDiffError()
    products_left = left["products"]
    products_right = copy(right["products"])

    if len(products_left) != len(products_right):
        diffs.append(
            DifferentProductSetsError(
                f"# products differ: {len(products_left)} != {len(products_right)}"
            )
        )

    for left_idx, left_product in enumerate(products_left):
        left_product_name = components(left_product["name"])
        for right_idx, right_product in enumerate(products_right):
            if components(right_product["name"]) != left_product_name:
                continue
            try:
                compare_product_membership(left_product, right_product)
            except MultiDiffError as compare_diffs:
                diffs.extend(compare_diffs)
            products_right.remove(right_product)
            break
        else:
            diffs.append(
                DifferentProductSetsError(f"Left has {len(products_left)} left over")
            )

    if len(products_right) != 0:
        diffs.append(
            DifferentProductSetsError(f"Right has {len(products_right)} left over")
        )

    if diffs:
        raise diffs


def compare_product_membership(left, right):
    """Compare membership between products

    Parameters
    ---------
    left, right : dict
        Two, individual, association products to compare

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    diffs = MultiDiffError()
    if len(right["members"]) != len(left["members"]):
        diffs.append(
            MemberMismatchError(
                "Product Member length differs:"
                " Left Product {left_product_name} len {left_len} !=  "
                " Right Product {right_product_name} len {right_len}"
                "".format(
                    left_product_name=left["name"],
                    left_len=len(left["members"]),
                    right_product_name=right["name"],
                    right_len=len(right["members"]),
                )
            )
        )

    members_right = copy(right["members"])
    for left_member in left["members"]:
        for right_member in members_right:
            if left_member["expname"] != right_member["expname"]:
                continue

            if left_member["exptype"] != right_member["exptype"]:
                diffs.append(
                    MemberMismatchError(
                        "Left {left_expname}:{left_exptype} != Right"
                        " {right_expname}:{right_exptype}".format(
                            left_expname=left_member["expname"],
                            left_exptype=left_member["exptype"],
                            right_expname=right_member["expname"],
                            right_exptype=right_member["exptype"],
                        )
                    )
                )

            members_right.remove(right_member)
            break
        else:
            diffs.append(
                MemberMismatchError(
                    f"Left {left_member['expname']}:{left_member['exptype']} has no"
                    " counterpart in right"
                )
            )

    if len(members_right) != 0:
        diffs.append(
            MemberMismatchError(
                "Right has {len_over} unaccounted for members starting with"
                " {right_expname}:{right_exptype}".format(
                    len_over=len(members_right),
                    right_expname=members_right[0]["expname"],
                    right_exptype=members_right[0]["exptype"],
                )
            )
        )

    if diffs:
        raise diffs


def components(s):
    """split string into its components"""
    return set(re.split("[_-]", s))


def separate_products(asn):
    """Separate products into individual associations

    Parameters
    ----------
    asn: `Association`
        The association to split

    Returns
    -------
    separated: [`Association`[, ...]]
        The list of separated associations
    """
    separated = []
    for product in asn["products"]:
        new_asn = copy(asn)
        new_asn["products"] = [product]
        separated.append(new_asn)
    return separated


def get_product_names(asns):
    """Return product names from associations and flag duplicates

    Parameters
    ----------
    asns: [`Association`[, ...]]

    Returns
    -------
    product_names, duplicates: set(str[, ...]), [str[,...]]
        2-tuple consisting of the set of product names and the list of duplicates.
    """
    product_names = [asn["products"][0]["name"] for asn in asns]

    dups = [name for name, count in Counter(product_names).items() if count > 1]
    if dups:
        logger.debug("Duplicate product names: %s", dups)

    return set(product_names), dups



"""Unit-like tests related to FluxStep"""

import astropy.units as u
import numpy as np
import pytest
from roman_datamodels import datamodels, maker_utils

from romancal.datamodels.container import ModelContainer
from romancal.flux import FluxStep
from romancal.flux.flux_step import LV2_UNITS


@pytest.mark.parametrize(
    "attr, factor",
    [
        ("data", 1),
        ("err", 1),
        ("var_rnoise", 2),
        ("var_poisson", 2),
        ("var_flat", 2),
    ],
)
def test_attributes(flux_step, attr, factor):
    """Test that the attribute has been scaled by the right factor"""
    original, result = flux_step
    c_unit = 1.0 / LV2_UNITS

    # Handle difference between just a single image and a list.
    if isinstance(original, datamodels.ImageModel):
        original_list = [original]
        result_list = [result]
    else:
        original_list = original
        result_list = result

    for original_model, result_model in zip(original_list, result_list):
        c_mj = original_model.meta.photometry.conversion_megajanskys
        scale = (c_mj * c_unit) ** factor
        original_value = getattr(original_model, attr)
        result_value = getattr(result_model, attr)

        assert np.allclose(original_value * scale, result_value)


# ########
# Fixtures
# ########
@pytest.fixture(scope="module", params=["input_imagemodel", "input_modelcontainer"])
def flux_step(request):
    """Execute FluxStep on given input

    Parameters
    ----------
    input : str, `roman_datamodels.datamodels.DataModel`, or `~romancal.datamodels.container.ModelContainer`

    Returns
    -------
    original, result : DataModel or ModelContainer, DataModel or ModelContainer
    """
    input = request.getfixturevalue(request.param)

    # Copy input because flux operates in-place
    original = input.copy()

    # Perform step
    result = FluxStep.call(input)

    # That's all folks
    return original, result


@pytest.fixture(scope="module")
def image_model():
    """Product a basic ImageModel"""
    # Create a random image and specify a conversion.
    rng = np.random.default_rng()
    shape = (10, 10)
    image_model = maker_utils.mk_datamodel(datamodels.ImageModel, shape=shape)
    image_model.data = u.Quantity(
        rng.poisson(2.5, size=shape).astype(np.float32),
        LV2_UNITS,
        dtype=np.float32,
    )
    image_model.var_rnoise = u.Quantity(
        rng.normal(1, 0.05, size=shape).astype(np.float32),
        LV2_UNITS**2,
        dtype=np.float32,
    )
    image_model.var_poisson = u.Quantity(
        rng.poisson(1, size=shape).astype(np.float32),
        LV2_UNITS**2,
        dtype=np.float32,
    )
    image_model.var_flat = u.Quantity(
        rng.uniform(0, 1, size=shape).astype(np.float32),
        LV2_UNITS**2,
        dtype=np.float32,
    )
    image_model.meta.photometry.conversion_megajanskys = 2.0 * u.MJy / u.sr

    return image_model


@pytest.fixture(scope="module")
def input_imagemodel(image_model):
    """Provide a single ImageModel"""

    # First just setup the basic model
    return image_model.copy()


@pytest.fixture(scope="module")
def input_modelcontainer(image_model):
    """Provide a ModelContainer"""
    # Create and return a ModelContainer
    image_model1 = image_model.copy()
    image_model2 = image_model.copy()
    image_model2.meta.photometry.conversion_megajanskys = 0.5 * u.MJy / u.sr
    container = ModelContainer([image_model1, image_model2])
    return container



"""
Module to detect sources using image segmentation.
"""

import logging
import math
import warnings

from astropy.convolution import convolve
from astropy.units import Quantity
from astropy.utils.exceptions import AstropyUserWarning
from photutils.segmentation import SourceFinder, make_2dgaussian_kernel
from photutils.utils.exceptions import NoDetectionsWarning

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def convolve_data(data, kernel_fwhm, size=None, mask=None):
    """
    Convolve the background-subtracted model image with a Gaussian
    kernel.

    Parameters
    ----------
    data : 2D `astropy.units.Quantity`
        The input 2D Quantity array. The data is assumed to be
        background subtracted.
    kernel_fwhm : float
        The FWHM of the Gaussian kernel.
    size : int, optional
        The size of the kernel array. Default is 3 times the kernel
        FWHM.
    mask : 2D `numpy.ndarray`, optional
        A boolean mask with the same shape as the input data. True
        values indicate masked pixels.

    Returns
    -------
    convolved_data : `numpy.ndarray`
        The convolved data array.
    """
    if not isinstance(data, Quantity):
        raise ValueError("Input model must be a Quantity array.")

    size = math.ceil(kernel_fwhm * 3)
    size = size + 1 if size % 2 == 0 else size  # make size be odd
    kernel = make_2dgaussian_kernel(kernel_fwhm, size=size)  # normalized to 1

    with warnings.catch_warnings():
        # suppress warnings caused by large masked areas
        warnings.simplefilter("ignore", AstropyUserWarning)
        return convolve(data, kernel, mask=mask)


def make_segmentation_image(
    convolved_data, snr_threshold, npixels, bkg_rms, deblend=False, mask=None
):
    """
    Make a segmentation image from a model image.

    Parameters
    ----------
    convolved_data : 2D `numpy.ndarray`
        The background-subtracted convolved data array.
    snr_threshold : float
        The per-pixel signal-to-noise ratio threshold for detection.
    npixels : int
        The number of connected pixels required to define a source.
    bkg_rms : 2D `numpy.ndarray`
        The background RMS array.
    deblend : bool, optional
        Whether to deblend overlapping sources. Default is False.
    mask : 2D `numpy.ndarray`, optional
        A boolean mask with the same shape as the input data. True
        values indicate masked pixels.

    Returns
    -------
    segment_img : `SegmentationImage`
        The segmentation image.
    """
    with warnings.catch_warnings():
        # suppress NoDetectionsWarning from photutils
        warnings.filterwarnings("ignore", category=NoDetectionsWarning)

        finder = SourceFinder(npixels, deblend=deblend)
        threshold = snr_threshold * bkg_rms
        segment_img = finder(convolved_data, threshold, mask=mask)

        if segment_img is None:
            nsources = 0
        else:
            nsources = segment_img.nlabels
        log.info(f"Detected {nsources} sources")

    return segment_img



import logging

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

# Guide star mode exposure types
GUIDER_LIST = [
    "WFI_WIM_ACQ",
    "WFI_WIM_TRACK",
    "WFI_WSM_ACQ1",
    "WFI_WSM_ACQ2",
    "WFI_WSM_TRACK",
]


def do_dqinit(input_model, mask=None):
    """Check that the input model pixeldq attribute has the same dimensions as
    the image plane of the input model science data, call apply_dqinit, and update
    log and cal_step.  Changes input model in place.

    Parameters
    ----------
    input_model : input Roman datamodel
        The Roman datamodel to be data quality  corrected

    mask_model : Roman data model, or None
        The mask model to use in the data quality correction

    Returns
    -------
    output_model : Roman datamodel
        The data quality corrected Roman datamodel
    """

    # Change model in place
    output_model = input_model

    # Determine if mask is shapewise compatable with the input
    skip_step = False
    if input_model.meta.exposure.type in GUIDER_LIST:
        # Check to see if the shape of the mask data array
        # is not equal to the shape of the science data
        if input_model.dq.shape != mask.dq.shape:
            skip_step = True
    else:
        # Check to see if the shape of the mask data array
        # is not equal to the shape of the science data
        if input_model.pixeldq.shape != mask.dq.shape:
            skip_step = True

    # Apply or skip dq correction
    if skip_step:
        log.warning("Mask data array is not the same shape as the science data")
        log.warning("Step will be skipped")
        output_model.meta.cal_step.dq_init = "SKIPPED"
    else:
        output_model = apply_dqinit(input_model, mask)
        output_model.meta.cal_step.dq_init = "COMPLETE"

    return output_model


def apply_dqinit(science, mask):
    """Apply data quality mask to the dq or pixeldq arrays.

    Extended summary
    ----------------
    The dq attribute of the MASK model is bitwise OR'd with the
    pixeldq (or dq) attribute of the input model.

    Parameters
    ----------
    science : Roman data model
        input science data model

    mask : Roman data model
        data quality mask data model
    """

    # Set model-specific data quality in output
    if science.meta.exposure.type in GUIDER_LIST:
        science.dq |= mask.dq
    else:
        science.pixeldq |= mask.dq

    return science



"""Tests for asn_from_list"""

import pytest

from romancal.associations import Association, AssociationRegistry, load_asn
from romancal.associations.asn_from_list import Main, asn_from_list
from romancal.associations.exceptions import AssociationNotValidError


def test_base_association():
    """Create the simplest of associations"""
    items = ["a", "b", "c"]
    asn = asn_from_list(items, rule=Association)
    assert asn["asn_rule"] == "Association"
    assert asn["asn_type"] == "None"
    assert asn["members"] == items


def test_base_roundtrip():
    """Write/read created base association"""
    items = ["a", "b", "c"]
    asn = asn_from_list(items, rule=Association)
    _, serialized = asn.dump()
    reloaded = load_asn(serialized, registry=None)
    assert asn["asn_rule"] == reloaded["asn_rule"]
    assert asn["asn_type"] == reloaded["asn_type"]
    assert asn["members"] == reloaded["members"]


def test_default_simple():
    """Default ELPP association"""
    product_name = "test_product"
    items = ["a", "b", "c"]
    asn = asn_from_list(items, product_name=product_name)
    assert asn["asn_rule"] == "DMS_ELPP_Base"
    assert asn["asn_type"] == "None"
    assert len(asn["products"]) == 1
    product = asn["products"][0]
    assert product["name"] == product_name
    assert len(product["members"]) == len(items)
    for member in product["members"]:
        assert member["expname"] in items
        assert member["exptype"] == "science"


def test_default_with_type():
    """ELPP association with types specified"""
    product_name = "test_product"
    items = {"a": "science", "b": "target_acq", "c": "somethingelse"}
    asn = asn_from_list(
        list(items.items()), product_name=product_name, with_exptype=True
    )
    assert asn["asn_rule"] == "DMS_ELPP_Base"
    assert asn["asn_type"] == "None"
    assert len(asn["products"]) == 1
    product = asn["products"][0]
    assert product["name"] == product_name
    assert len(product["members"]) == len(items)
    for member in product["members"]:
        assert member["expname"] in items
        assert member["exptype"] == items[member["expname"]]


def test_default_fail():
    """Test default DMS_ELPP_Base fail

    A product name needs to be included, but is not.
    """
    items = ["a"]
    with pytest.raises(AssociationNotValidError):
        _ = asn_from_list(items)


def test_default_roundtrip():
    """Create/Write/Read a ELPP association"""
    product_name = "test_product"
    items = {"a": "science", "b": "target_acq", "c": "somethingelse"}
    asn = asn_from_list(
        [(item, type_) for item, type_ in items.items()],
        product_name=product_name,
        with_exptype=True,
    )
    _, serialized = asn.dump()
    reloaded = load_asn(serialized)
    assert asn["asn_rule"] == reloaded["asn_rule"]
    assert asn["asn_type"] == reloaded["asn_type"]
    assert len(asn["products"]) == len(reloaded["products"])


def test_cmdline_fails():
    """Exercise the command line interface"""

    # No arguments
    with pytest.raises(SystemExit):
        Main([])

    # Only the association file argument
    with pytest.raises(SystemExit):
        Main(["-o", "test_asn.json"])


@pytest.mark.parametrize("format", ["json", "yaml"])
def test_cmdline_success(format, tmp_path):
    """Create ELPP associations in different formats"""
    path = tmp_path / "test_asn.json"
    product_name = "test_product"
    inlist = ["a", "b", "c"]
    args = ["-o", str(path), "--product-name", product_name, "--format", format]
    args = args + inlist
    Main(args)
    with path.open() as fp:
        asn = load_asn(fp, format=format)
    assert len(asn["products"]) == 1
    assert asn["products"][0]["name"] == product_name
    members = asn["products"][0]["members"]
    expnames = [member["expname"] for member in members]
    assert inlist == expnames


def test_cmdline_change_rules(tmp_path):
    """Command line change the rule"""
    rule = "Asn_Lv2Image"
    path = tmp_path / "test_asn.json"
    inlist = ["a", "b", "c"]
    args = [
        "-o",
        str(path),
        "-r",
        rule,
        "--product-name",
        "test",
    ]
    args = args + inlist
    Main(args)
    with path.open() as fp:
        asn = load_asn(fp, registry=AssociationRegistry(include_bases=True))
    # assert inlist == asn['members']
    assert inlist[0] == asn["products"][0]["members"][0]["expname"]
    assert inlist[1] == asn["products"][0]["members"][1]["expname"]
    assert inlist[2] == asn["products"][0]["members"][2]["expname"]


def test_api_list():
    """Test api call with simple list"""
    product_name = "test_product"
    inlist = ["a", "b", "c"]

    asn = asn_from_list(inlist, product_name=product_name)
    assert len(asn["products"]) == 1
    assert asn["products"][0]["name"] == product_name
    members = asn["products"][0]["members"]
    expnames = [member["expname"] for member in members]
    assert inlist == expnames


def test_api_with_type():
    """Test api call with type tuple"""
    product_name = "test_product"
    inlist = [("a", "science"), ("b", "psf"), ("c", "science")]

    asn = asn_from_list(inlist, product_name=product_name, with_exptype=True)
    assert len(asn["products"]) == 1
    assert asn["products"][0]["name"] == product_name
    members = asn["products"][0]["members"]
    members_dict = {member["expname"]: member["exptype"] for member in members}
    for name, type_ in inlist:
        assert members_dict[name] == type_



"""Load an Association from a file or object"""

from inspect import isclass

from .association import Association
from .registry import AssociationRegistry


def load_asn(
    serialized,
    format=None,
    first=True,
    validate=True,
    registry=AssociationRegistry,
    **kwargs,
):
    """Load an Association from a file or object

    Parameters
    ----------
    serialized : object
        The serialized form of the association.

    format : str or None
        The format to force. If None, try all available.

    validate : bool
        Validate against the class' defined schema, if any.

    first : bool
        A serialization potentially matches many rules.
        Only return the first succesful load.

    registry : AssociationRegistry or None
        The `AssociationRegistry` to use.
        If None, no registry is used.
        Can be passed just a registry class instead of instance.

    kwargs : dict
        Other arguments to pass to the `load` methods defined
        in the `Association.IORegistry`

    Returns
    -------
    The Association object

    Raises
    ------
    AssociationNotValidError
        Cannot create or validate the association.

    Notes
    -----
    The `serialized` object can be in any format
    supported by the registered I/O routines. For example, for
    `json` and `yaml` formats, the input can be either a string or
    a file object containing the string.

    If no registry is specified, the default `Association.load`
    method is used.
    """
    if registry is None:
        return Association.load(serialized, format=format, validate=validate)

    if isclass(registry):
        registry = registry()
    return registry.load(
        serialized, format=format, first=first, validate=validate, **kwargs
    )



"""Reprocessing List"""

from collections import deque
from enum import Enum
from functools import reduce

__all__ = [
    "ListCategory",
    "ProcessList",
    "ProcessItem",
    "ProcessQueue",
    "ProcessQueueSorted",
]


class ListCategory(Enum):
    """Categories in the list"""

    RULES = 0
    BOTH = 1
    EXISTING = 2
    NONSCIENCE = 3


class ProcessItem:
    """Items to be processed

    Parameters
    ----------
    obj : object
        The object to make a `ProcessItem`.
        Objects must be equatable.
    """

    def __init__(self, obj):
        self.obj = obj

    @classmethod
    def to_process_items(cls, iterable):
        """Iterable to convert a list to ProcessItem's

        Parameters
        ----------
        iterable : iterable
            A source of objects to convert

        Returns
        -------
        An iterable where the object has been
        converted to a `ProcessItem`
        """
        for obj in iterable:
            yield cls(obj)

    def __hash__(self):
        try:
            hash_value = self.obj.__hash__()
        except (AttributeError, TypeError):
            hash_value = hash(repr(self))
        return hash_value

    def __eq__(self, other):
        try:
            equality = self.obj == other.obj
        except AttributeError:
            equality = self.__hash__() == other.__hash__()
        return equality


class ProcessList:
    """A Process list

    Parameters
    ----------
    items : [item[, ...]]
        The list of items to process

    rules : [Association[, ...]]
        List of rules to process the items against.

    work_over : int
        What the reprocessing should work on:
        - `ProcessList.EXISTING`:   Only existing associations
        - `ProcessList.RULES`:      Only on the rules to create new associations
        - `ProcessList.BOTH`:       Compare to both existing and rules
        - `ProcessList.NONSCIENCE`: Only on non-science items

    only_on_match : bool
        Only use this object if the overall condition
        is True.

    trigger_constraints : [Constraint[,...]]
        The constraints that created the ProcessList

    trigger_rules : [Association[,...]]
        The association rules that created the ProcessList
    """

    _str_attrs = (
        "rules",
        "work_over",
        "only_on_match",
        "trigger_constraints",
        "trigger_rules",
    )

    def __init__(
        self,
        items=None,
        rules=None,
        work_over=ListCategory.BOTH,
        only_on_match=False,
        trigger_constraints=None,
        trigger_rules=None,
    ):
        self.items = items
        self.rules = rules
        self.work_over = work_over
        self.only_on_match = only_on_match
        self.trigger_constraints = (
            set(trigger_constraints) if trigger_constraints else set()
        )
        self.trigger_rules = set(trigger_rules) if trigger_rules else set()

    @property
    def hash(self):
        """Create a unique hash"""
        return (tuple(self.rules), self.work_over, self.only_on_match)

    def update(self, process_list, full=False):
        """Update with information from ProcessList

        Attributes from `process_list` are added to self's attributes. If `not
        full`, the attributes `rules`, 'work_over`, and `only_on_match` are not
        taken.

        Note that if `full`, destructive action will occur with respect to
        `work_over` and `only_on_match`.

        Parameters
        ----------
        process_list : ProcessList
            The source process list to absorb.

        full : bool
            Include the hash attributes `rules`, `work_over`, and `only_on_match`.
        """
        self.items += process_list.items
        self.trigger_constraints.update(process_list.trigger_constraints)
        self.trigger_rules.update(process_list.trigger_rules)
        if full:
            self.rules += process_list.rules
            self.work_over = process_list.work_over
            self.only_on_match = process_list.only_on_match

    def __str__(self):
        result = "{}(n_items: {}, {})".format(
            self.__class__.__name__,
            len(self.items),
            {str_attr: getattr(self, str_attr) for str_attr in self._str_attrs},
        )
        return result


class ProcessQueue(deque):
    """Make a deque iterable and mutable"""

    def __iter__(self):
        while True:
            try:
                yield self.popleft()
            except (IndexError, StopIteration):
                break


class ProcessListQueue:
    """First-In-First-Out queue of ProcessLists

    Parameters
    ----------
    init : [ProcessList[,...]] or None
        List of ProcessLists to put on the queue.
    """

    def __init__(self, init=None):
        self._queue = {}
        if init is not None:
            self.extend(init)

    def append(self, process_list):
        """Add ProcessList to queue"""
        plhash = process_list.hash
        if plhash not in self._queue:
            self._queue[plhash] = process_list
        else:
            self._queue[plhash].update(process_list)

    def extend(self, iterable):
        """Add objects if not already in the queue"""
        for process_list in iterable:
            self.append(process_list)

    def items(self):
        """Return list generator of all items"""
        for plhash in self._queue:
            yield from self._queue[plhash].items

    def popleft(self):
        """Pop the first-in object"""
        plhash = next(iter(self._queue))
        process_list = self._queue[plhash]
        del self._queue[plhash]
        return process_list

    def __len__(self):
        return len(self._queue)

    def __iter__(self):
        while True:
            try:
                yield self.popleft()
            except (IndexError, StopIteration):
                break

    def __str__(self):
        result = (
            f"{self.__class__.__name__}: rulesets {len(self)} items"
            f" {len(list(self.items()))}"
        )
        return result


class ProcessQueueSorted:
    """Sort ProcessItem based on work_over

    `ProcessList`s are handled in order of `RULES`, `BOTH`,
    `EXISTING`, and `NONSCIENCE`.

    Parameters
    ----------
    init : [ProcessList[,...]]
        List of `ProcessList` to start the queue with.
    """

    def __init__(self, init=None):
        self.queues = {
            list_category: ProcessListQueue() for list_category in ListCategory
        }

        if init is not None:
            self.extend(init)

    def extend(self, process_lists):
        """Add the list of process items to their appropriate queues"""
        for process_list in process_lists:
            self.queues[process_list.work_over].append(process_list)

    def __iter__(self):
        """Return the queues in order"""
        while len(self) > 0:
            for category in ListCategory:
                for process_list in self.queues[category]:
                    yield process_list
                    break
                else:
                    continue
                break

    def __len__(self):
        return reduce(lambda x, y: x + len(y), self.queues.values(), 0)

    def __str__(self):
        result = f"{self.__class__.__name__}:"
        for queue in self.queues:
            result += f"\n\tQueue {queue}: {self.queues[queue]}"
        return result


def workover_filter(process_list, work_over):
    """Determine and modify workover of input process list

    Parameters
    ---------
    process_list : ProcessList
        The process list under consideration

    work_over : ListCategory
        The ListCategory to compare against.

    Returns
    -------
    process_list : ProcessList or None
        The input process_list with work_over modified.
        None if the process list should not be continued.
    """
    result = process_list
    if process_list.work_over in [ListCategory.RULES, ListCategory.BOTH]:
        if work_over in [ListCategory.RULES, ListCategory.BOTH]:
            result.work_over = ListCategory.BOTH
        else:
            result.work_over = work_over
    else:
        if work_over not in [ListCategory.RULES, ListCategory.BOTH]:
            result = None
    return result



"""
The `skystatistics` module provides statistics computation class used by
:py:func:`~romancal.skymatch.skymatch.match`
and :py:class:`~romancal.skymatch.skyimage.SkyImage`.
"""

from copy import deepcopy

# THIRD PARTY
from stsci.imagestats import ImageStats

__all__ = ["SkyStats"]


class SkyStats:
    """
    This is a superclass build on top of
    :py:class:`stsci.imagestats.ImageStats`. Compared to
    :py:class:`stsci.imagestats.ImageStats`, `SkyStats` has
    "persistent settings" in the sense that object's parameters need to be
    set once and these settings will be applied to all subsequent
    computations on different data.
    """

    def __init__(
        self,
        skystat="mean",
        lower=None,
        upper=None,
        nclip=5,
        lsig=4.0,
        usig=4.0,
        binwidth=0.1,
        **kwargs,
    ):
        """Initializes the SkyStats object.

        Parameters
        -----------
        skystat : {'mode', 'median', 'mode', 'midpt'}, optional
            Sets the statistics that will be returned by `~SkyStats.calc_sky`.
            The following statistics are supported: 'mean', 'mode', 'midpt',
            and 'median'. First three statistics have the same meaning as in
            `stsdas.toolbox.imgtools.gstatistics <http://stsdas.stsci.edu/\
cgi-bin/gethelp.cgi?gstatistics>`_
            while 'median' will compute the median of the distribution.

        lower : float, None, optional
            Lower limit of usable pixel values for computing the sky.
            This value should be specified in the units of the input image(s).

        upper : float, None, optional
            Upper limit of usable pixel values for computing the sky.
            This value should be specified in the units of the input image(s).

        nclip : int, optional
            A non-negative number of clipping iterations to use when computing
            the sky value.

        lsig : float, optional
            Lower clipping limit, in sigma, used when computing the sky value.

        usig : float, optional
            Upper clipping limit, in sigma, used when computing the sky value.

        binwidth : float, optional
            Bin width, in sigma, used to sample the distribution of pixel
            brightness values in order to compute the sky background
            statistics.

        kwargs : dict
            A dictionary of optional arguments to be passed to `ImageStats`.

        """
        self.npix = None
        self.skyval = None

        self._fields = ",".join(["npix", skystat])

        self._kwargs = deepcopy(kwargs)
        if "fields" in self._kwargs:
            del self._kwargs["fields"]
        if "image" in self._kwargs:
            del self._kwargs["image"]
        self._kwargs["lower"] = lower
        self._kwargs["upper"] = upper
        self._kwargs["nclip"] = nclip
        self._kwargs["lsig"] = lsig
        self._kwargs["usig"] = usig
        self._kwargs["binwidth"] = binwidth

        self._skystat = {
            "mean": self._extract_mean,
            "mode": self._extract_mode,
            "median": self._extract_median,
            "midpt": self._extract_midpt,
        }[skystat]

    def _extract_mean(self, imstat):
        return imstat.mean

    def _extract_median(self, imstat):
        return imstat.median

    def _extract_mode(self, imstat):
        return imstat.mode

    def _extract_midpt(self, imstat):
        return imstat.midpt

    def calc_sky(self, data):
        """Computes statistics on data.

        Parameters
        -----------
        data : numpy.ndarray
            A numpy array of values for which the statistics needs to be computed.

        Returns
        --------
        statistics : tuple
            A tuple of two values: (`skyvalue`, `npix`), where `skyvalue`
            is the statistics specified by the `skystat` parameter during
            the initialization of the `SkyStats` object and `npix` is the
            number of pixels used in computing the statistics reported
            in `skyvalue`.

        """
        imstat = ImageStats(image=data.value, fields=self._fields, **(self._kwargs))
        stat = self._skystat(imstat)  # dict or scalar

        # re-attach units:
        if hasattr(stat, "__len__"):
            self.skyval = {k: value * data.unit for k, value in stat.items()}
        else:
            self.skyval = stat * data.unit

        self.npix = imstat.npix
        return self.skyval, self.npix

    def __call__(self, data):
        return self.calc_sky(data)



import json
import os
from io import StringIO
from pathlib import Path

import pytest
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils as utils

from romancal.datamodels.container import ModelContainer, make_file_with_index


@pytest.fixture()
def test_data_dir():
    return Path.joinpath(Path(__file__).parent, "data")


def create_asn_file(tmp_path, products: [] = None):
    asn_content = """
        {
            "asn_type": "None",
            "asn_rule": "DMS_ELPP_Base",
            "version_id": null,
            "code_version": "0.9.1.dev28+ge987cc9.d20230106",
            "degraded_status": "No known degraded exposures in association.",
            "program": "noprogram",
            "constraints": "No constraints",
            "asn_id": "a3001",
            "target": "none",
            "asn_pool": "test_pool_name",
            "products": [
                {
                    "name": "files.asdf",
                    "members": [
                        {
                            "expname": "img_1.asdf",
                            "exptype": "science",
                            "tweakreg_catalog": "img_1_catalog.cat"
                        },
                        {
                            "expname": "img_2.asdf",
                            "exptype": "science"
                        }
                    ]
                }
            ]
        }
"""
    if products is not None:
        temp_json = json.loads(asn_content)
        temp_json["products"] = products
        asn_content = json.dumps(temp_json)

    asn_file_path = str(tmp_path / "sample_asn.json")
    asn_file = StringIO()
    asn_file.write(asn_content)
    with open(asn_file_path, mode="w") as f:
        print(asn_file.getvalue(), file=f)

    return asn_file_path


@pytest.fixture()
def setup_list_of_l2_files():
    def _setup_list_of_l2_files(n, obj_type, tmp_path):
        """
        Generate a list of `n` ASDF files (and their corresponding path) or datamodels.

        Parameters
        ----------
        n : int
            The number of ASDF files or datamodels to be generated.
        obj_type : str
            The type of object to be generated. Allowed values: "asdf" or "datamodel".
        tmp_path : _type_
            The dir path where the generated ASDF files will be temporarily saved to.

        Returns
        -------
        list
            A list containing either the full path to an ASDF file or datamodels.
        """
        number_of_files_to_create = n
        type_of_returned_object = obj_type

        result_list = []
        for i in range(number_of_files_to_create):
            filepath = (
                tmp_path
                / f"test_model_container_input_as_list_of_filepaths_{i:02}.asdf"
            )
            # create an ASDF file with an L2 model
            utils.mk_level2_image(filepath=filepath, shape=(100, 100))
            if type_of_returned_object == "asdf":
                # append filepath to filepath list
                result_list.append(str(filepath))
            elif type_of_returned_object == "datamodel":
                # parse ASDF file as RDM
                datamodel = rdm.open(str(filepath))
                # update filename
                datamodel.meta["filename"] = filepath
                # append datamodel to datamodel list
                result_list.append(datamodel)

        return result_list

    return _setup_list_of_l2_files


def test_model_container_init_with_modelcontainer_instance():
    img = utils.mk_level2_image()
    m = rdm.ImageModel(img)
    mc1 = ModelContainer([m])

    # initialize with an instance of ModelContainer
    mc2 = ModelContainer(mc1)

    assert isinstance(mc2, ModelContainer)
    assert all(isinstance(x, rdm.DataModel) for x in mc1)
    assert len(mc1) == len(mc2)


@pytest.mark.parametrize("n, obj_type", [(3, "asdf"), (2, "datamodel")])
def test_model_container_init_path_to_asdf_or_datamodels(
    n, obj_type, tmp_path, request
):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list)

    assert all(x in filepath_list for x in mc._models)


def test_model_container_init_with_path_to_asn_file(tmp_path):
    # create ASDF files with L2 datamodel with custom tweakreg_catalog file
    utils.mk_level2_image(filepath=tmp_path / "img_1.asdf")
    utils.mk_level2_image(filepath=tmp_path / "img_2.asdf")
    # create ASN file that points to the ASDF files
    asn_filepath = create_asn_file(tmp_path)
    mc = ModelContainer(asn_filepath)

    assert all(hasattr(x.meta, "asn") for x in mc)


@pytest.mark.parametrize(
    "input_object",
    [
        "invalid_object",
        "",
        [1, 2, 3],
        ModelContainer(),
        Path(),
    ],
)
def test_imagemodel_init_error(input_object):
    with pytest.raises(Exception) as e:
        ModelContainer(input_object)

    assert e.type == TypeError


@pytest.mark.parametrize("n, obj_type", [(4, "asdf"), (5, "datamodel")])
def test_imagemodel_slice_n_dice(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    # provide filepath list as input to ModelContainer
    mc = ModelContainer(filepath_list)

    x1 = mc[:]
    x2 = mc[0]
    x3 = mc[2:]
    x4 = mc[-2]
    x5 = mc[-2:]

    assert isinstance(x1, list)
    assert len(x1) == len(filepath_list)

    assert isinstance(x2, rdm.ImageModel)

    assert isinstance(x3, list)
    assert len(x3) == n - 2

    assert isinstance(x4, rdm.ImageModel)

    assert isinstance(x5, list)
    assert len(x5) == 2


def test_imagemodel_set_item(setup_list_of_l2_files, tmp_path):
    filepath_list = setup_list_of_l2_files(4, "datamodel", tmp_path)
    # provide filepath list as input to ModelContainer
    mc1 = ModelContainer(filepath_list[:2])
    mc2 = ModelContainer(filepath_list[-2:])

    mc1[0] = mc2[-2]
    mc1[1] = mc2[-1]

    assert all(id(l) == id(r) for l, r in zip(mc1[:], mc2[:]))


@pytest.mark.parametrize(
    "n, obj_type, input_object",
    [
        (2, "datamodel", "invalid_object"),
        (2, "datamodel", ""),
        (2, "datamodel", [1, 2, 3]),
        (2, "datamodel", ModelContainer()),
    ],
)
def test_imagemodel_set_item_error(n, obj_type, input_object, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    # provide filepath list as input to ModelContainer
    mc = ModelContainer(filepath_list)

    with pytest.raises(Exception) as e:
        mc[0] = input_object

    assert e.type == ValueError


@pytest.mark.parametrize("n, obj_type", [(3, "datamodel")])
def test_model_container_insert(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list[:2])

    mc.insert(1, filepath_list[-1])

    assert len(mc) == n
    assert id(mc[1]) == id(filepath_list[-1])


@pytest.mark.parametrize("n, obj_type", [(3, "datamodel")])
def test_model_container_insert_error(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list)

    with pytest.raises(Exception) as e:
        # try to insert a ModelContainer
        mc.insert(1, ModelContainer())

    assert e.type == ValueError


@pytest.mark.parametrize("n, obj_type", [(3, "datamodel")])
def test_model_container_append(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list[:2])

    mc.append(filepath_list[-1])

    assert len(mc) == n
    assert id(mc[-1]) == id(filepath_list[-1])


@pytest.mark.parametrize("n, obj_type", [(3, "datamodel")])
def test_model_container_append_error(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list)

    with pytest.raises(Exception) as e:
        # try to append a ModelContainer
        mc.append(ModelContainer())

    assert e.type == ValueError


@pytest.mark.parametrize("n, obj_type", [(4, "datamodel")])
def test_model_container_extend(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc1 = ModelContainer(filepath_list[:2])
    mc2 = ModelContainer(filepath_list[-2:])

    mc1.extend(mc2)

    assert len(mc1) == n
    assert all(id(l) == id(r) for l, r in zip(mc1[-2:], filepath_list[-2:]))


@pytest.mark.parametrize(
    "n, obj_type, input_object",
    [
        (3, "datamodel", ["trying_to_sneak_in", 1, 2, 3]),
        (3, "datamodel", ""),
        (3, "datamodel", "trying_to_sneak_in"),
    ],
)
def test_model_container_extend_error(n, obj_type, input_object, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list)

    with pytest.raises(Exception) as e:
        # try to insert a string
        mc.extend(input_object)

    assert e.type == ValueError


@pytest.mark.parametrize("n, obj_type", [(3, "datamodel")])
def test_model_container_pop_last_item(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list)

    mc.pop()

    assert len(mc) == n - 1
    assert filepath_list[-1] not in mc[:]
    assert all(id(l) == id(r) for l, r in zip(mc[:], filepath_list[:]))


@pytest.mark.parametrize("n, obj_type", [(3, "datamodel")])
def test_model_container_pop_with_index(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )
    mc = ModelContainer(filepath_list)

    index_to_be_removed = 1
    mc.pop(index_to_be_removed)

    assert len(mc) == n - 1
    assert filepath_list[index_to_be_removed] not in mc[:]


@pytest.mark.parametrize("n, obj_type", [(2, "asdf"), (3, "datamodel")])
def test_model_container_copy(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )

    mc = ModelContainer(filepath_list)
    mc_copy_dict = mc.copy()

    mc_dict = mc.__dict__
    mc_copy_dict = mc_copy_dict.__dict__

    assert id(mc_dict) != id(mc_copy_dict)
    assert all(x in mc_dict for x in mc_copy_dict)
    assert all(
        type(o) == type(c) for o, c in zip(mc_copy_dict.values(), mc_dict.values())
    )


@pytest.mark.parametrize("n, obj_type", [(2, "asdf"), (3, "datamodel")])
def test_get_crds_parameters(n, obj_type, tmp_path, request):
    filepath_list = request.getfixturevalue("setup_list_of_l2_files")(
        n, obj_type, tmp_path
    )

    assert isinstance(ModelContainer(filepath_list).get_crds_parameters(), dict)


def test_get_crds_parameters_empty():
    crds_param = ModelContainer().get_crds_parameters()

    assert isinstance(crds_param, dict)
    assert len(crds_param) == 0


def test_close_all_datamodels(setup_list_of_l2_files, tmp_path):
    filepath_list = setup_list_of_l2_files(3, "datamodel", tmp_path)

    mc = ModelContainer(filepath_list)

    mc.close()

    assert all(x._asdf._closed for x in mc)


def test_add_tweakreg_catalog_attribute_from_asn(tmp_path):
    # create ASDF files with L2 datamodel
    utils.mk_level2_image(filepath=tmp_path / "img_1.asdf")
    utils.mk_level2_image(filepath=tmp_path / "img_2.asdf")
    # create ASN file that points to the ASDF files
    asn_filepath = create_asn_file(tmp_path)
    mc = ModelContainer(asn_filepath)

    assert hasattr(mc[0].meta, "tweakreg_catalog")


def test_models_grouped(setup_list_of_l2_files, tmp_path):
    filepath_list = setup_list_of_l2_files(3, "datamodel", tmp_path)

    mc = ModelContainer(filepath_list)

    generated_group = mc.models_grouped
    generated_group_id = {x.meta.group_id for x in list(generated_group)[0]}
    generated_group_members = list(list(generated_group)[0])

    unique_exposure_parameters = [
        "program",
        "observation",
        "visit",
        "visit_file_group",
        "visit_file_sequence",
        "visit_file_activity",
        "exposure",
    ]
    params = [
        str(getattr(mc[0].meta.observation, param))
        for param in unique_exposure_parameters
    ]
    expected_group_id = "roman" + "_".join(
        ["".join(params[:3]), "".join(params[3:6]), params[6]]
    )

    assert all(hasattr(x.meta, "group_id") for x in mc)
    assert generated_group_id.pop() == expected_group_id
    assert all(id(l) == id(r) for l, r in zip(generated_group_members, mc._models))


def test_merge_tree():
    mc = ModelContainer()

    a = {
        "a_k1": "a_v1",
        "a_k2": "a_v2",
        "a_k3": "a_v3",
        "a_k4": "a_v4",
        "a_k5": "a_v5",
    }
    b = {
        "b_k1": "b_v1",
        "b_k2": "b_v2",
        "b_k3": "b_v3",
        "b_k4": "b_v4",
        "b_k5": "b_v5",
    }

    mc.merge_tree(a, b)

    assert all(x in a for x in b)
    assert all(x in a.values() for x in b.values())


@pytest.mark.parametrize("asn_filename", ["detector_asn.json", "detectorFOV_asn.json"])
def test_parse_asn_files_properly(asn_filename, test_data_dir):
    # instantiate a MC without reading/loading/saving datamodels
    mc = ModelContainer(
        test_data_dir / f"{asn_filename}",
        return_open=False,
        save_open=False,
    )

    with open(test_data_dir / f"{asn_filename}") as f:
        json_content = json.load(f)
    # extract expname from json file
    expname_list = [
        x["expname"].split("/")[-1] for x in json_content["products"][0]["members"]
    ]

    assert len(mc) == len(json_content["products"][0]["members"])
    assert mc.asn_table_name == f"{asn_filename}"
    assert all(x.split("/")[-1] in expname_list for x in mc)


@pytest.mark.parametrize(
    "path, dir_path, save_model_func, output_suffix",
    [(None, None, None, None), (None, None, None, "output")],
)
def test_model_container_save(
    path,
    dir_path,
    save_model_func,
    setup_list_of_l2_files,
    output_suffix,
    tmp_path,
):
    filepath_list = setup_list_of_l2_files(3, "datamodel", tmp_path)

    mc = ModelContainer(filepath_list)

    output_paths = mc.save(
        path=path,
        dir_path=dir_path,
        save_model_func=save_model_func,
        output_suffix=output_suffix,
    )

    assert all(Path(x).exists() for x in output_paths)

    # clean up
    [os.remove(filename) for filename in output_paths]


@pytest.mark.parametrize(
    "filename, idx, expected_filename_result",
    [("file.asdf", None, "file.asdf"), ("file.asdf", 0, "file0.asdf")],
)
def test_make_file_with_index(filename, idx, expected_filename_result, tmp_path):
    filepath = str(tmp_path / filename)
    result = make_file_with_index(file_path=filepath, idx=idx)

    assert result == str(tmp_path / expected_filename_result)


def test_modelcontainer_works_properly_with_context_manager(tmp_path):
    """Test that ModelContainer works correctly with context manager."""

    products = [
        {
            "name": "files.asdf",
            "members": [
                {"expname": "img_1.asdf", "exptype": "science"},
                {"expname": "img_2.asdf", "exptype": "science"},
                {"expname": "img_3.asdf", "exptype": "science"},
            ],
        }
    ]

    asn_filepath = create_asn_file(tmp_path, products=products)

    with ModelContainer(asn_filepath, return_open=False, save_open=False) as asn1:
        assert type(asn1) is ModelContainer


@pytest.mark.parametrize(
    "exc_type, exc_val, exc_tb, test_id",
    [
        (None, None, None, "no_exception"),
        (ValueError, ValueError("Test error"), None, "value_error"),
        (TypeError, TypeError("Type error"), None, "type_error"),
    ],
)
def test_modelcontainer_exits_properly_on_exception(exc_type, exc_val, exc_tb, test_id):
    """Test that ModelContainer exits properly with context manager."""
    mc = ModelContainer()

    result = mc.__exit__(exc_type, exc_val, exc_tb)

    assert result is False



"""Project default for pytest"""

import inspect
import json
import os
import tempfile
from io import StringIO

import pytest
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling.models import Shift
from gwcs import coordinate_frames as cf
from gwcs import wcs
from roman_datamodels import datamodels as rdm
from roman_datamodels import maker_utils

from romancal.assign_wcs import pointing

collect_ignore = ["lib/dqflags.py"]


@pytest.fixture
def mk_tmp_dirs():
    """Create a set of temporary directorys and change to one of them."""
    tmp_current_path = tempfile.mkdtemp()
    tmp_data_path = tempfile.mkdtemp()
    tmp_config_path = tempfile.mkdtemp()

    old_path = os.getcwd()
    try:
        os.chdir(tmp_current_path)
        yield (tmp_current_path, tmp_data_path, tmp_config_path)
    finally:
        os.chdir(old_path)


@pytest.fixture
def slow(request):
    """Setup slow fixture for tests to identify if --slow
    has been specified
    """
    return request.config.getoption("--slow")


@pytest.fixture(scope="function")
def function_jail(tmp_path):
    """Perform test in a pristine temporary working directory."""
    old_dir = os.getcwd()
    os.chdir(tmp_path)
    try:
        yield str(tmp_path)
    finally:
        os.chdir(old_dir)


@pytest.fixture(scope="module")
def module_jail(request, tmp_path_factory):
    """Run test in a pristine temporary working directory, scoped to module.

    This fixture is the same as _jail in ci_watson, but scoped to module
    instead of function.  This allows a fixture using it to produce files in a
    temporary directory, and then have the tests access them.
    """
    old_dir = os.getcwd()
    path = request.module.__name__.split(".")[-1]
    if request._parent_request.fixturename is not None:
        path = path + "_" + request._parent_request.fixturename
    newpath = tmp_path_factory.mktemp(path)
    os.chdir(str(newpath))
    yield newpath
    os.chdir(old_dir)


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    terminal_reporter = config.pluginmanager.getplugin("terminalreporter")
    config.pluginmanager.register(
        TestDescriptionPlugin(terminal_reporter), "testdescription"
    )


class TestDescriptionPlugin:
    """Pytest plugin to print the test docstring when `pytest -vv` is used.

    This plug-in was added to support JWST instrument team testing and
    reporting for the JWST calibration pipeline.
    """

    def __init__(self, terminal_reporter):
        self.terminal_reporter = terminal_reporter
        self.desc = None

    def pytest_runtest_protocol(self, item):
        try:
            # Get the docstring for the test
            self.desc = inspect.getdoc(item.obj)
        except AttributeError:
            self.desc = None

    @pytest.hookimpl(hookwrapper=True, tryfirst=True)
    def pytest_runtest_logstart(self, nodeid, location):
        # When run as `pytest` or `pytest -v`, no change in behavior
        if self.terminal_reporter.verbosity <= 1:
            yield
        # When run as `pytest -vv`, `pytest -vvv`, etc, print the test docstring
        else:
            self.terminal_reporter.write("\n")
            yield
            if self.desc:
                self.terminal_reporter.write(f"\n{self.desc} ")


@pytest.fixture(scope="function")
def create_mock_asn_file():
    def _create_asn_file(tmp_path: str, members_mapping: dict = None) -> str:
        """
        Create a mock association file with the provided members mapping.

        Parameters
        ----------
        tmp_path : str
            Path to the temporary directory to store the association file.
        members_mapping : list, optional
            Mapping of members to include in the association file (default: None).

        Returns
        -------
        str
            Path to the created association file.
        """
        asn_content = """
            {
                "asn_type": "None",
                "asn_rule": "DMS_ELPP_Base",
                "version_id": null,
                "code_version": "0.9.1.dev28+ge987cc9.d20230106",
                "degraded_status": "No known degraded exposures in association.",
                "program": "noprogram",
                "constraints": "No constraints",
                "asn_id": "a3001",
                "target": "none",
                "asn_pool": "test_pool_name",
                "products": [
                    {
                        "name": "files.asdf",
                        "members": [
                            {
                                "expname": "img_1.asdf",
                                "exptype": "science"
                            },
                            {
                                "expname": "img_2.asdf",
                                "exptype": "science"
                            }
                        ]
                    }
                ]
            }
        """
        if members_mapping:
            asn_dict = json.loads(asn_content)
            asn_dict["products"][0]["members"] = []
            for x in members_mapping:
                asn_dict["products"][0]["members"].append(x)
            asn_content = json.dumps(asn_dict)

        asn_file_path = str(tmp_path / "sample_asn.json")
        asn_file = StringIO()
        asn_file.write(asn_content)
        with open(asn_file_path, mode="w") as f:
            print(asn_file.getvalue(), file=f)

        return asn_file_path

    return _create_asn_file


def _create_wcs(input_dm, shift_1=0, shift_2=0):
    """
    Create a basic WCS object (with optional shift) and
    append it to the input_dm.meta attribute.

    The WCS object will have a pipeline with the required
    steps to validate against the TweakReg pipeline.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    shift_1 : int, optional
        Shift to be applied in the direction of the first axis, by default 0
    shift_2 : int, optional
        Shift to be applied in the direction of the second axis, by default 0
    """

    shape = input_dm.data.shape

    # create necessary transformations
    distortion = Shift(-shift_1) & Shift(-shift_2)
    distortion.bounding_box = ((-0.5, shape[-1] + 0.5), (-0.5, shape[-2] + 0.5))
    tel2sky = pointing.v23tosky(input_dm)

    # create required frames
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    v2v3 = cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    world = cf.CelestialFrame(reference_frame=coord.ICRS(), name="world")

    # create pipeline
    pipeline = [
        wcs.Step(detector, distortion),
        wcs.Step(v2v3, tel2sky),
        wcs.Step(world, None),
    ]

    wcs_obj = wcs.WCS(pipeline)

    input_dm.meta["wcs"] = wcs_obj


@pytest.fixture
def base_image():
    """
    Create a base image with a realistic WCSInfo and a WCS.

    Notes
    -----
    The size of the image needs to be relatively large in order for
    the source catalog step to find a reasonable number of sources in the image.

    shift_1 and shift_2 (units in pixel) are used to shift the WCS projection plane.
    """

    def _base_image(shift_1=0, shift_2=0):
        l2 = maker_utils.mk_level2_image(shape=(100, 100))
        l2_im = rdm.ImageModel(l2)
        _create_wcs(l2_im)
        l2_im.meta.wcsinfo.vparity = -1
        return l2_im

    return _base_image



"""Constraints
"""

import abc
import collections
import logging
import re
from copy import deepcopy
from itertools import chain

from ..pool import PoolRow
from .process_list import ListCategory, ProcessList
from .utilities import evaluate, getattr_from_list, is_iterable

__all__ = [
    "AttrConstraint",
    "Constraint",
    "ConstraintTrue",
    "SimpleConstraint",
]

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class SimpleConstraintABC(abc.ABC):
    """Simple Constraint ABC

    Parameters
    ----------
    init : dict
        dict where the key:value pairs define
        the following parameters

    value : object or None
        Value that must be matched.

    name : str or None
        Option name for constraint

    **kwargs : key:value pairs
        Other initialization parameters

    Attributes
    ----------
    matched : bool
        Last call to `check_and_set`
    """

    # Attributes to show in the string representation.
    _str_attrs = ("name", "value")

    def __init__(self, init=None, value=None, name=None, **kwargs):
        # Defined attributes
        self.value = value
        self.name = name
        self.matched = False

        if init is not None:
            self.__dict__.update(init)
        else:
            self.__dict__.update(kwargs)

    @abc.abstractmethod
    def check_and_set(self, item):
        """Check and set the constraint

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        self.matched = True
        return self.matched, []

    @property
    def dup_names(
        self,
    ):  #  -> dict[str, list[typing.Union[SimpleConstraint, Constraint]]]
        """Return dictionary of constraints with duplicate names

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Returns
        -------
        dups : {str: [constraint[,...]][,...]}
            Returns a mapping between the duplicated name
            and all the constraints that define that name.
        """
        return {}

    @property
    def id(self):
        """Return identifyer for the constraint

        Returns
        -------
        id : str
            The identifyer
        """
        return f"{self.__class__.__name__}:{self.name}"

    def copy(self):
        """Copy ourselves"""
        return deepcopy(self)

    def get_all_attr(
        self, attribute: str
    ):  # -> list[tuple[SimpleConstraint, typing.Any]]:
        """Return the specified attribute

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Parameters
        ----------
        attribute : str
            The attribute to retrieve

        Returns
        -------
        [(self, value)] : [(SimpleConstraint, object)]
            The value of the attribute in a tuple. If there is no attribute,
            an empty tuple is returned.
        """
        value = getattr(self, attribute)
        if value is not None:
            return [(self, value)]
        return []

    # Make iterable to work with `Constraint`.
    # Since this is a leaf, simple return ourselves.
    def __iter__(self):
        yield self

    def __repr__(self):
        result = f"{self.__class__.__name__}({str(self.__dict__)})"
        return result

    def __str__(self):
        attrs = {str_attr: getattr(self, str_attr) for str_attr in self._str_attrs}
        return f"{self.__class__.__name__}({attrs})"


class ConstraintTrue(SimpleConstraintABC):
    """Always return True"""

    def check_and_set(self, item):
        return super().check_and_set(item)


class SimpleConstraint(SimpleConstraintABC):
    """A basic constraint

    Parameters
    ----------
    init : dict
        dict where the key:value pairs define
        the following parameters

    value : object or None
        Value that must be matched.
        If None, any retrieved value will match.

    sources : func(item) or None
        Function taking `item` as argument used to
        retrieve a value to check against.
        If None, the item itself is used as the value.

    force_unique : bool
        If the constraint is satisfied, reset `value`
        to the value of the source.

    test : function
        The test function for the constraint.
        Takes two arguments:

            - constraint
            - object to compare against.

        Returns a boolean.
        Default is `SimpleConstraint.eq`

    name : str or None
        Option name for constraint

    reprocess_on_match : bool
        Reprocess the item if the constraint is satisfied.

    reprocess_on_fail : bool
        Reprocess the item if the constraint is not satisfied.

    work_over : ListCategory.[BOTH, EXISTING, RULES]
        The condition on which this constraint should operate.

    reprocess_rules : [rule[,..]] or None
        List of rules to be applied to.
        If None, calling function will determine the ruleset.
        If empty, [], all rules will be used.

    Attributes
    ----------
    All `Parameters` are also `Attributes`

    Examples
    --------

    Create a constraint where the attribute `attr` of an object
    matches the value `my_value`:

    >>> c = SimpleConstraint(value='my_value')
    >>> print(c)
    SimpleConstraint({'name': None, 'value': 'my_value'})

    To check a constraint, call `check_and_set`. A successful match
    will return a tuple of `True` and a reprocess list.
    >>> item = 'my_value'
    >>> c.check_and_set(item)
    (True, [])

    If it doesn't match, `False` will be returned.
    >>> bad_item = 'not_my_value'
    >>> c.check_and_set(bad_item)
    (False, [])

    A `SimpleConstraint` can also be initialized by a `dict`
    of the relevant parameters:
    >>> init = {'value': 'my_value'}
    >>> c = SimpleConstraint(init)
    >>> print(c)
    SimpleConstraint({'name': None, 'value': 'my_value'})

    If the value to check is `None`, the `SimpleConstraint` will
    successfully match whatever object given. However, a new `SimpleConstraint`
    will be returned where the `value` is now set to whatever the attribute
    was of the object.
    >>> c = SimpleConstraint(value=None)
    >>> matched, reprocess = c.check_and_set(item)
    >>> print(c)
    SimpleConstraint({'name': None, 'value': 'my_value'})

    This behavior can be overridden by the `force_unique` parameter:
    >>> c = SimpleConstraint(value=None, force_unique=False)
    >>> matched, reprocess = c.check_and_set(item)
    >>> print(c)
    SimpleConstraint({'name': None, 'value': None})

    """

    def __init__(
        self,
        init=None,
        sources=None,
        force_unique=True,
        test=None,
        reprocess_on_match=False,
        reprocess_on_fail=False,
        work_over=ListCategory.BOTH,
        reprocess_rules=None,
        **kwargs,
    ):
        # Defined attributes
        self.sources = sources
        self.force_unique = force_unique
        self.test = test
        self.reprocess_on_match = reprocess_on_match
        self.reprocess_on_fail = reprocess_on_fail
        self.work_over = work_over
        self.reprocess_rules = reprocess_rules
        super().__init__(init=init, **kwargs)

        # Give defaults some real meaning.
        if self.sources is None:
            self.sources = lambda item: item
        if test is None:
            self.test = self.eq

    def check_and_set(self, item):
        """Check and set the constraint

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        source_value = self.sources(item)

        satisfied = True
        if self.value is not None:
            satisfied = self.test(self.value, source_value)
        self.matched = satisfied

        if self.matched:
            if self.force_unique:
                self.value = source_value

        # Determine reprocessing
        reprocess = []
        if (self.matched and self.reprocess_on_match) or (
            not self.matched and self.reprocess_on_fail
        ):
            reprocess.append(
                ProcessList(
                    items=[item],
                    work_over=self.work_over,
                    rules=self.reprocess_rules,
                    trigger_constraints=[self.id],
                )
            )

        return self.matched, reprocess

    def eq(self, value1, value2):
        """True if constraint.value and item are equal."""
        return value1 == value2


class AttrConstraint(SimpleConstraintABC):
    """Test attribute of an item

    Parameters
    ----------
    sources : [str[,...]]
        List of attributes to query

    value : str, function or None
        The value to check for. If None and
        `force_unique`, any value in the first
        available source will become the value.
        If function, the function takes no arguments
        and returns a string.

    evaluate : bool
        Evaluate the item's value before checking condition.

    force_reprocess : ListCategory.state or False
        Add item back onto the reprocess list using
        the specified `~jwst.associations.ProcessList` work over state.

    force_unique : bool
        If the initial value is `None` or a list of possible values,
        the constraint will be modified to be the value first matched.

    invalid_values : [str[,...]]
        List of values that are invalid in an item.
        Will cause a non-match.

    name : str or None
        Name of the constraint.

    only_on_match : bool
        If `force_reprocess`, only do the reprocess
        if the entire constraint is satisfied.

    onlyif : function
        Boolean function that takes `item` as argument.
        If True, the rest of the condition is checked. Otherwise
        return as a matched condition

    required : bool
        One of the sources must exist. Otherwise,
        return as a matched constraint.

    Attributes
    ----------
    found_values : set(str[,...])
        Set of actual found values for this condition.

    matched : bool
        Last result of `check_and_set`

    """

    # Attributes to show in the string representation.
    _str_attrs = ("name", "sources", "value")

    def __init__(
        self,
        init=None,
        sources=None,
        evaluate=False,
        force_reprocess=False,
        force_undefined=False,
        force_unique=True,
        invalid_values=None,
        only_on_match=False,
        onlyif=None,
        required=True,
        **kwargs,
    ):
        # Attributes
        self.sources = sources
        self.evaluate = evaluate
        self.force_reprocess = force_reprocess
        self.force_undefined = force_undefined
        self.force_unique = force_unique
        self.invalid_values = invalid_values
        self.only_on_match = only_on_match
        self.onlyif = onlyif
        self.required = required
        super().__init__(init=init, **kwargs)

        # Give some defaults real meaning.
        if invalid_values is None:
            self.invalid_values = []
        if onlyif is None:
            self.onlyif = lambda item: True

        # Haven't actually matched anything yet.
        self.found_values = set()
        self.matched = False

    def check_and_set(self, item):
        """Check and set constraints based on item

        Parameters
        ----------
        item : dict
            The item to check on.

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        reprocess = []

        # Only perform check on specified `onlyif` condition
        if not self.onlyif(item):
            if self.force_reprocess:
                reprocess.append(
                    ProcessList(
                        items=[item],
                        work_over=self.force_reprocess,
                        only_on_match=self.only_on_match,
                        trigger_constraints=[self.id],
                    )
                )
            self.matched = True
            return self.matched, reprocess

        # Get the condition information.
        try:
            source, value = getattr_from_list(
                item, self.sources, invalid_values=self.invalid_values
            )
        except KeyError:
            if self.required and not self.force_undefined:
                self.matched = False
                return self.matched, reprocess
            else:
                self.matched = True
                return self.matched, reprocess
        else:
            if self.force_undefined:
                self.matched = False
                return self.matched, reprocess

        evaled = value
        if self.evaluate:
            evaled = evaluate(value)

        # If the constraint has no value to check against, and given
        # value evaluates to a list, the item must be duplicated,
        # with each value from its list, and all the new items reprocessed.
        # Otherwise, the value is the value to set the constraint by.
        if self.value is None:
            if is_iterable(evaled):
                reprocess.append(reprocess_multivalue(item, source, evaled, self))
                self.matched = False
                return self.matched, reprocess
            value = str(evaled)

        # Else, the constraint does have a value. Check against it.
        else:
            if callable(self.value):
                match_value = self.value()
            else:
                match_value = self.value
            if not is_iterable(evaled):
                evaled = [evaled]
            matched = []
            for evaled_item in evaled:
                value = str(evaled_item)
                if meets_conditions(value, match_value):
                    matched.append(evaled_item)

            # If the condition is not matched, leave now.
            if not matched:
                self.matched = False
                return self.matched, reprocess

            # Matches were found. The first match becomes
            # the value for the current constraint.
            # The rest get reprocessed.
            value = str(matched[0])
            if matched[1:]:
                reprocess.append(reprocess_multivalue(item, source, matched[1:], self))

        # At this point, the constraint has passed.
        # Fix the conditions.
        escaped_value = re.escape(value)
        self.found_values.add(escaped_value)
        if self.force_unique:
            self.value = escaped_value
            self.sources = [source]
            self.force_unique = False

        # If required to reprocess, add to the reprocess list.
        if self.force_reprocess:
            reprocess.append(
                ProcessList(
                    items=[item],
                    work_over=self.force_reprocess,
                    only_on_match=self.only_on_match,
                    trigger_constraints=[self.id],
                )
            )

        # That's all folks
        self.matched = True
        return self.matched, reprocess


class Constraint:
    """Constraint that is made up of SimpleConstraints

    Parameters
    ----------
    init : object or [object[,...]]
        A single object or list of objects where the
        objects are as follows.
        - SimpleConstraint or subclass
        - Constraint

    reduce : function
        A reduction function with signature `x(iterable)`
        where `iterable` is the `components` list. Returns
        boolean indicating state of the components.
        Default value is `Constraint.all`

    name : str or None
        Optional name for constraint.

    reprocess_on_match : bool
        Reprocess the item if the constraint is satisfied.

    reprocess_on_fail : bool
        Reprocess the item if the constraint is not satisfied.

    work_over : ListCategory.[BOTH, EXISTING, RULES]
        The condition on which this constraint should operate.

    reprocess_rules : [rule[,..]] or None
        List of rules to be applied to.
        If None, calling function will determine the ruleset.
        If empty, [], all rules will be used.

    Attributes
    ----------
    constraints : [Constraint[,...]]
        List of `Constraint` or `SimpleConstraint` that
        make this constraint.

    matched : bool
        Result of the last `check_and_set`

    reduce : function
        A reduction function with signature `x(iterable)`
        where `iterable` is the `components` list. Returns
        boolean indicating state of the components.
        Predefined functions are:
        - `all`: True if all components return True
        - `any`: True if any component returns True

    Notes
    -----
    Named constraints can be accessed directly through indexing:

    >>> c = Constraint(SimpleConstraint(name='simple', value='a_value'))
    >>> c['simple']  # doctest: +SKIP
    SimpleConstraint({'sources': <function SimpleConstraint.__init__.<locals>.<lambda> at 0x7f8be05f5730>,
                      'force_unique': True,
                      'test': <bound method SimpleConstraint.eq of SimpleConstraint({...})>,
                      'reprocess_on_match': False,
                      'reprocess_on_fail': False,
                      'work_over': 1,
                      'reprocess_rules': None,
                      'value': 'a_value',
                      'name': 'simple',
                      'matched': False})
    """  # noqa: E501

    def __init__(
        self,
        init=None,
        reduce=None,
        name=None,
        reprocess_on_match=False,
        reprocess_on_fail=False,
        work_over=ListCategory.BOTH,
        reprocess_rules=None,
    ):
        self.constraints = []

        # Initialize from named parameters
        self.reduce = reduce
        self.name = name
        self.reprocess_on_match = reprocess_on_match
        self.reprocess_on_fail = reprocess_on_fail
        self.work_over = work_over
        self.reprocess_rules = reprocess_rules

        # Initialize from a structure.
        if init is None:
            pass
        elif isinstance(init, list):
            self.constraints = init
        elif isinstance(init, Constraint):
            self.reduce = init.reduce
            self.name = init.name
            self.reprocess_on_match = init.reprocess_on_match
            self.reprocess_on_fail = init.reprocess_on_fail
            self.work_over = init.work_over
            self.reprocess_rules = init.reprocess_rules
            self.constraints = deepcopy(init.constraints)
        elif isinstance(init, SimpleConstraintABC):
            self.constraints = [init]
        else:
            raise TypeError(
                "Invalid initialization value type {}.\nValid types are"
                " `SimpleConstraint`, `Constraint`,\nor subclass.".format(type(init))
            )

        # Give some defaults real meaning.
        self.matched = False
        if self.reduce is None:
            self.reduce = self.all

    @property
    def dup_names(
        self,
    ):  # -> dict[str, list[typing.Union[SimpleConstraint, Constraint]]]:
        """Return dictionary of constraints with duplicate names

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Returns
        -------
        dups : {str: [constraint[,...]][,...]}
            Returns a mapping between the duplicated name
            and all the constraints that define that name.
        """
        attrs = self.get_all_attr("name")
        constraints, names = zip(*attrs)
        dups = [name for name, count in collections.Counter(names).items() if count > 1]
        result = collections.defaultdict(list)
        for name, constraint in zip(names, constraints):
            if name in dups:
                result[name].append(constraint)

        # Turn off the defaultdict factory.
        result.default_factory = None
        return result

    @property
    def id(self):
        """Return identifyer for the constraint

        Returns
        -------
        id : str
            The identifyer
        """
        return f"{self.__class__.__name__}:{self.name}"

    def append(self, constraint):
        """Append a new constraint"""
        self.constraints.append(constraint)

    def check_and_set(self, item, work_over=ListCategory.BOTH):
        """Check and set the constraint

        Returns
        -------
        success, reprocess : bool, [~jwst.associations.ProcessList[,...]]
            Returns 2-tuple of

                - success : True if check is successful.
                - List of `~jwst.associations.ProcessList`.
        """
        if work_over not in (self.work_over, ListCategory.BOTH):
            return False, []

        # Do we have positive?
        self.matched, reprocess = self.reduce(item, self.constraints)

        # Determine reprocessing
        if (self.matched and self.reprocess_on_match) or (
            not self.matched and self.reprocess_on_fail
        ):
            reprocess.append(
                [
                    ProcessList(
                        items=[item],
                        work_over=self.work_over,
                        rules=self.reprocess_rules,
                        trigger_constraints=[self.id],
                    )
                ]
            )

        return self.matched, list(chain(*reprocess))

    def copy(self):
        """Copy ourselves"""
        return deepcopy(self)

    def get_all_attr(
        self, attribute: str
    ):  # -> list[tuple[typing.Union[SimpleConstraint, Constraint], typing.Any]]:
        """Return the specified attribute

        This method is meant to be overridden by classes
        that need to traverse a list of constraints.

        Parameters
        ----------
        attribute : str
            The attribute to retrieve

        Returns
        -------
        result : [(SimpleConstraint or Constraint, object)[,...]]
            The list of values of the attribute in a tuple. If there is no attribute,
            an empty tuple is returned.

        Raises
        ------
        AttributeError
            If the attribute is not found.
        """
        result = []
        value = getattr(self, attribute)
        if value is not None:
            result = [(self, value)]
        for constraint in self.constraints:
            result.extend(constraint.get_all_attr(attribute))

        return result

    @staticmethod
    def all(item, constraints):
        """Return positive only if all results are positive."""

        # If there are no constraints, there is nothing to match.
        # Result is false.
        if len(constraints) == 0:
            return False, []

        # Find all negatives. Note first negative
        # that requires reprocessing and how many
        # negatives do not.
        all_match = True
        negative_reprocess = None
        to_reprocess = []
        for constraint in constraints:
            match, reprocess = constraint.check_and_set(item)

            if match:
                if all_match:
                    to_reprocess.append(reprocess)
            else:
                all_match = False

                # If not match and no reprocessing, then fail
                # completely. However, if there is reprocessing, take
                # the first one. Continue to check to ensure
                # there is no further complete fail.
                if len(reprocess) == 0:
                    negative_reprocess = None
                    break
                elif negative_reprocess is None:
                    negative_reprocess = [reprocess]

        if not all_match:
            if negative_reprocess is not None:
                to_reprocess = negative_reprocess
            else:
                to_reprocess = []

        return all_match, to_reprocess

    @staticmethod
    def any(item, constraints):
        """Return the first successful constraint."""
        # If there are no constraints, there is nothing to match.
        # Result is false.
        if len(constraints) == 0:
            return False, []

        to_reprocess = []
        for constraint in constraints:
            match, reprocess = constraint.check_and_set(item)
            if match:
                to_reprocess = [reprocess]
                break
            to_reprocess.append(reprocess)
        return match, to_reprocess

    @staticmethod
    def notany(item, constraints):
        """True if none of the constraints match"""
        match, to_reprocess = Constraint.any(item, constraints)
        return not match, to_reprocess

    @staticmethod
    def notall(item, constraints):
        """True if not all of the constraints match"""
        match, to_reprocess = Constraint.all(item, constraints)
        return not match, to_reprocess

    def __delitem__(self, key):
        """Not implemented"""
        raise NotImplementedError("Cannot delete a constraint by index.")

    # Make iterable
    def __iter__(self):
        yield from chain(*map(iter, self.constraints))

    # Index implementation
    def __getitem__(self, key):
        """Retrieve a named constraint"""
        for constraint in self.constraints:
            name = getattr(constraint, "name", None)
            if name is not None and name == key:
                return constraint
            try:
                found = constraint[key]
            except (KeyError, TypeError):
                pass
            else:
                return found
        raise KeyError(f"Constraint {key} not found")

    def __repr__(self):
        result = "{}(name={}).{}([{}])".format(
            self.__class__.__name__,
            str(getattr(self, "name", None)),
            str(self.reduce.__name__),
            "".join([repr(constraint) for constraint in self.constraints]),
        )
        return result

    def __setitem__(self, key, value):
        """Not implemented"""
        raise NotImplementedError("Cannot set constraints by index.")

    def __str__(self):
        result = "\n".join(
            [str(constraint) for constraint in self if constraint.name is not None]
        )
        return result


# ---------
# Utilities
# ---------
def meets_conditions(value, conditions):
    """Check whether value meets any of the provided conditions

    Parameters
    ----------
    values : str
        The value to be check with.

    condition : regex,
        Regular expressions to match against.

    Returns
    -------
    True if any condition is meant.
    """

    if not is_iterable(conditions):
        conditions = [conditions]
    for condition in conditions:
        condition = "".join(["^", condition, "$"])
        match = re.match(condition, value, flags=re.IGNORECASE)
        if match:
            return True
    return False


def reprocess_multivalue(item, source, values, constraint):
    """Reprocess items that have a list of values

    Parameters
    ----------
    item : dict
        The item.

    source : str
        The attribute which has the multi-values.

    values : list
        The list of values

    constraint : Constraint
        The constraint which is triggering the reprocessing.

    Returns
    -------
    process_list : ProcessList
        The process list to put on the reprocess queue
    """
    reprocess_items = []
    for value in values:
        new_item = PoolRow(item)
        new_item[source] = str(value)
        reprocess_items.append(new_item)
    process_list = ProcessList(
        items=reprocess_items, trigger_constraints=[constraint.id]
    )
    return process_list



import numpy as np
from astropy.modeling.models import RotationSequence3D, Scale
from gwcs.geometry import CartesianToSpherical, SphericalToCartesian


def v23tosky(input_model, wrap_v2_at=180, wrap_lon_at=360):
    """Create the transform from telescope to sky.

    The transform is defined with a reference point in a Frame
    associated tih the telescope (V2, V3) in arcsec, the corresponding
    reference poiont on sky (RA_REF, DEC_REF) in deg, and the position angle
    at the center of the aperture, ROLL_REF in deg.

    Parameters
    ----------
    input_model : `roman_datamodels.WfiImage`
        Roman imaging exposure data model.
    wrap_v2_at : float
        At what angle to wrap V2. [deg]
    wrap_lon_at : float
        At what angle to wrap logitude. [deg]

    Returns
    -------
    model : `astropy.modeling.Model`
        The transform from V2,V3 to sky.
    """
    v2_ref = input_model.meta.wcsinfo.v2_ref / 3600
    v3_ref = input_model.meta.wcsinfo.v3_ref / 3600
    roll_ref = input_model.meta.wcsinfo.roll_ref
    ra_ref = input_model.meta.wcsinfo.ra_ref
    dec_ref = input_model.meta.wcsinfo.dec_ref

    angles = np.array([v2_ref, -v3_ref, roll_ref, dec_ref, -ra_ref])
    axes = "zyxyz"
    rot = RotationSequence3D(angles, axes_order=axes)

    # The sky rotation expects values in deg.
    # This should be removed when models work with quantities.
    model = (
        (Scale(1 / 3600) & Scale(1 / 3600))
        | SphericalToCartesian(wrap_lon_at=wrap_v2_at)
        | rot
        | CartesianToSpherical(wrap_lon_at=wrap_lon_at)
    )
    model.name = "v23tosky"
    return model



"""Test the Registry"""

import pytest

from romancal.associations.lib.keyvalue_registry import KeyValueRegistry


@pytest.mark.skip(reason="Need to impliment")
def test_registry_match(full_pool_rules):
    """Test the match method"""
    pool, rules, pool_fname = full_pool_rules

    assert len(rules.schemas) > 0
    matches = rules.match(pool[1])
    assert isinstance(matches, tuple)
    asns = matches[0]
    reprocess_list = matches[1]
    assert isinstance(asns, list)
    assert isinstance(reprocess_list, list)
    assert len(asns) >= 1


# Tests below for keyvalue_registry


def test_dict_like():
    """Test the basic to ensure similar to a dict"""
    data = {"a": 1, "b": 2}
    kvr = KeyValueRegistry(data)
    assert data.keys() == kvr.keys()
    assert set(data.values()) == set(kvr.values())

    assert kvr.get("a") == 1
    assert kvr.get("c", 3) == 3

    keys, values = zip(*kvr.items())
    assert set(data.keys()) == set(keys)
    assert set(data.values()) == set(values)

    kvr_copy = kvr.copy()
    assert set(kvr_copy) == set(kvr)

    assert kvr.pop("a") == 1
    assert kvr.pop("a", 3) == 3
    assert kvr.popitem() == ("b", 2)
    with pytest.raises(KeyError):
        kvr.popitem()

    kvr = KeyValueRegistry()
    kvr.update(data)
    assert data.keys() == kvr.keys()
    assert set(data.values()) == set(kvr.values())

    kvr.clear()
    assert len(kvr) == 0


def test_default():
    kvr = KeyValueRegistry(default={"a": 1})
    assert kvr.default == "a"
    assert kvr[None] == 1


def test_tuple():
    data = ("a", 1)
    kvr = KeyValueRegistry(data)
    assert kvr["a"] == 1

    kvr = KeyValueRegistry(default=data)
    assert kvr["a"] == 1
    assert kvr[None] == 1


def test_fn():
    def fn():
        return 1

    kvr = KeyValueRegistry(fn)
    assert kvr[fn.__name__] is fn

    kvr = KeyValueRegistry(default=fn)
    assert kvr[fn.__name__] is fn
    assert kvr[None] is fn


def test_decorator():
    kvr = KeyValueRegistry()

    @kvr
    def fn():
        return 1

    assert kvr[fn.__name__] is fn
    assert kvr[fn.__name__]() is fn()



from stpipe import entry_points
from stpipe.utilities import import_class

import romancal.pipeline
import romancal.step
from romancal.stpipe import RomanPipeline, RomanStep
from romancal.stpipe.integration import get_steps


def test_get_steps():
    tuples = get_steps()

    assert {t[0].split(".")[-1] for t in tuples} == set(
        romancal.step.__all__ + romancal.pipeline.__all__
    )

    for class_name, class_alias, is_pipeline in tuples:
        step_class = import_class(class_name)
        assert issubclass(step_class, RomanStep)
        assert step_class.class_alias == class_alias
        if is_pipeline:
            assert issubclass(step_class, RomanPipeline)


def test_entry_point():
    all_steps = entry_points.get_steps()
    roman_steps = [s for s in all_steps if s.package_name == "romancal"]
    tuples = {(s.class_name, s.class_alias, s.is_pipeline) for s in roman_steps}
    assert tuples == set(get_steps())



import numpy as np
import pytest
from asdf import AsdfFile
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling import models
from astropy.units import Quantity
from gwcs import WCS
from gwcs import coordinate_frames as cf
from roman_datamodels import datamodels, maker_utils

from romancal.resample import ResampleStep, resample_utils


class MockModel:
    def __init__(self, pixelarea_steradians, pixelarea_arcsecsq, pixel_scale_ratio):
        self.meta = MockMeta(
            pixelarea_steradians, pixelarea_arcsecsq, pixel_scale_ratio
        )


class MockMeta:
    def __init__(self, pixelarea_steradians, pixelarea_arcsecsq, pixel_scale_ratio):
        self.photometry = MockPhotometry(pixelarea_steradians, pixelarea_arcsecsq)
        self.resample = MockResample(pixel_scale_ratio)


class MockPhotometry:
    def __init__(self, pixelarea_steradians, pixelarea_arcsecsq):
        self.pixelarea_steradians = pixelarea_steradians
        self.pixelarea_arcsecsq = pixelarea_arcsecsq


class MockResample:
    def __init__(self, pixel_scale_ratio):
        self.pixel_scale_ratio = pixel_scale_ratio


class Mosaic:
    def __init__(self, fiducial_world, pscale, shape, filename, n_images):
        self.fiducial_world = fiducial_world
        self.pscale = pscale
        self.shape = shape
        self.filename = filename
        self.n_images = n_images

    def create_mosaic(self):
        """
        Create a dummy L3 datamodel given the coordinates of the fiducial point,
        a pixel scale, and the image shape and filename.

        Returns
        -------
        datamodels.MosaicModel
            An L3 MosaicModel datamodel.
        """
        l3 = maker_utils.mk_level3_mosaic(
            shape=self.shape,
            n_images=self.n_images,
        )
        # data from WFISim simulation of SCA #01
        l3.meta.filename = self.filename
        l3.meta["wcs"] = create_wcs_object_without_distortion(
            fiducial_world=self.fiducial_world,
            pscale=self.pscale,
            shape=self.shape,
        )
        l3.meta.wcs.forward_transform
        return datamodels.MosaicModel(l3)


def create_wcs_object_without_distortion(fiducial_world, pscale, shape, **kwargs):
    """
    Create a simple WCS object without either distortion or rotation.

    Parameters
    ----------
    fiducial_world : tuple
        A pair of values corresponding to the fiducial's world coordinate.
    pscale : tuple
        A pair of values corresponding to the pixel scale in each axis.
    shape : tuple
        A pair of values specifying the dimensions of the WCS object.

    Returns
    -------
    gwcs.WCS
        A gwcs.WCS object.
    """
    # components of the model
    shift = models.Shift() & models.Shift()
    affine = models.AffineTransformation2D(
        matrix=[[1, 0], [0, 1]], translation=[0, 0], name="pc_rotation_matrix"
    )
    scale = models.Scale(pscale[0]) & models.Scale(pscale[1])
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        fiducial_world[0],
        fiducial_world[1],
        180,
    )

    # transforms between frames
    # detector -> sky
    det2sky = shift | affine | scale | tan | celestial_rotation
    det2sky.name = "linear_transform"

    # frames
    detector_frame = cf.Frame2D(
        name="detector",
        axes_order=(0, 1),
        axes_names=("x", "y"),
        unit=(u.pix, u.pix),
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.FK5(), name="fk5", unit=(u.deg, u.deg)
    )

    pipeline = [
        (detector_frame, det2sky),
        (sky_frame, None),
    ]

    wcs_obj = WCS(pipeline)

    wcs_obj.bounding_box = kwargs.get(
        "bounding_box",
        (
            (-0.5, shape[-1] - 0.5),
            (-0.5, shape[-2] - 0.5),
        ),
    )

    wcs_obj.pixel_shape = shape[::-1]
    wcs_obj.array_shape = shape

    return wcs_obj


@pytest.fixture
def asdf_wcs_file():
    def _create_asdf_wcs_file(tmp_path):
        file_path = tmp_path / "wcs.asdf"
        wcs_data = create_wcs_object_without_distortion(
            (10, 0),
            (0.000031, 0.000031),
            (100, 100),
        )
        wcs = {"wcs": wcs_data}
        with AsdfFile(wcs) as af:
            af.write_to(file_path)
        return str(file_path)

    return _create_asdf_wcs_file


@pytest.mark.parametrize(
    "vals, name, min_vals, expected",
    [
        ([1, 2], "list1", None, [1, 2]),
        ([None, None], "list2", None, None),
        ([1, 2], "list4", [0, 0], [1, 2]),
    ],
)
def test_check_list_pars_valid(vals, name, min_vals, expected):
    step = ResampleStep()

    result = step._check_list_pars(vals, name, min_vals)
    assert result == expected


def test_load_custom_wcs_no_file():
    step = ResampleStep()
    result = step._load_custom_wcs(None, (512, 512))
    assert result is None


def test_load_custom_wcs_missing_output_shape(asdf_wcs_file):
    with pytest.raises(ValueError):
        step = ResampleStep()
        step._load_custom_wcs(asdf_wcs_file, None)


def test_load_custom_wcs_invalid_file(tmp_path):
    step = ResampleStep()
    invalid_file = tmp_path / "invalid.asdf"
    with open(invalid_file, "w") as f:
        f.write("invalid asdf file")

    with pytest.raises(ValueError):
        step._load_custom_wcs(str(invalid_file), (512, 512))


def test_load_custom_wcs_asdf_without_wcs_attribute(tmp_path):
    step = ResampleStep()
    file_path = tmp_path / "asdf_file.asdf"
    wcs = {}
    with AsdfFile(wcs) as af:
        af.write_to(file_path)

    with pytest.raises(KeyError):
        step._load_custom_wcs(str(file_path), (100, 100))


@pytest.mark.parametrize(
    "vals, name, min_vals, expected",
    [
        ([1, 2], "test", [0, 0], [1, 2]),
        ([None, None], "test", [0, 0], None),
        ([0, 0], "test", [0, 0], [0, 0]),
        ([1, 1], "test", [0, 0], [1, 1]),
        ([0, 1], "test", [0, 0], [0, 1]),
        ([1, 0], "test", [0, 0], [1, 0]),
    ],
)
def test_check_list_pars(vals, name, min_vals, expected):
    step = ResampleStep()

    result = step._check_list_pars(vals, name, min_vals)
    assert result == expected


@pytest.mark.parametrize(
    "vals, name, min_vals",
    [
        ([None, 2], "test", [0, 0]),
        ([1, None], "test", [0, 0]),
        ([1], "test", [0, 0]),
        ([1, 2, 3], "test", [0, 0]),
        ([None, None, None], "test", [0, 0]),
        ([1, 2], "test", [2, 2]),
    ],
)
def test_check_list_pars_exception(vals, name, min_vals):
    step = ResampleStep()

    with pytest.raises(ValueError):
        step._check_list_pars(vals, name, min_vals)


@pytest.mark.parametrize(
    """pixelarea_steradians, pixelarea_arcsecsq, pixel_scale_ratio,
    expected_steradians, expected_arcsecsq""",
    [
        # Happy path tests
        (1.0, 1.0, 2.0, 4.0, 4.0),
        (2.0, 2.0, 0.5, 0.5, 0.5),
        (0.0, 0.0, 2.0, 0.0, 0.0),
        (1.0, 1.0, 0.0, 0.0, 0.0),
        (None, 1.0, 2.0, None, 4.0),
        (1.0, None, 2.0, 4.0, None),
    ],
)
def test_update_phot_keywords(
    pixelarea_steradians,
    pixelarea_arcsecsq,
    pixel_scale_ratio,
    expected_steradians,
    expected_arcsecsq,
):
    step = ResampleStep()
    model = MockModel(pixelarea_steradians, pixelarea_arcsecsq, pixel_scale_ratio)

    step.update_phot_keywords(model)

    assert model.meta.photometry.pixelarea_steradians == expected_steradians
    assert model.meta.photometry.pixelarea_arcsecsq == expected_arcsecsq


@pytest.mark.parametrize(
    "good_bits, dq_array, expected_output",
    [
        (
            "~DO_NOT_USE+NON_SCIENCE",
            np.array([[0, 2**0, 0], [0, 2**13, 0], [0, 2**9, 0]]),
            np.array([[1, 0, 1], [1, 1, 1], [1, 0, 1]]),
        ),
        (
            "~513",
            np.array([[0, 2**0, 0], [2**13, 0, 0], [0, 0, 2**9]]),
            np.array([[1, 0, 1], [1, 1, 1], [1, 1, 0]]),
        ),
        (
            "~1+512",
            np.array([[2**13, 0, 0], [0, 0, 2**9], [0, 2**0, 0]]),
            np.array([[1, 1, 1], [1, 1, 0], [1, 0, 1]]),
        ),
        (
            "~1,512",
            np.array([[2**13, 2**0, 2**9], [0, 0, 0], [0, 0, 0]]),
            np.array([[1, 0, 0], [1, 1, 1], [1, 1, 1]]),
        ),
        (
            "LOW_QE+NONLINEAR",
            np.array([[2**13, 2**0, 2**16], [0, 0, 0], [0, 0, 0]]),
            np.array([[1, 0, 1], [1, 1, 1], [1, 1, 1]]),
        ),
        (
            "73728",
            np.array([[0, 0, 0], [0, 2**13, 2**0], [0, 2**16, 0]]),
            np.array([[1, 1, 1], [1, 1, 0], [1, 1, 1]]),
        ),
        (
            "8192+65536",
            np.array([[0, 0, 0], [0, 2**13, 0], [2**0, 2**16, 0]]),
            np.array([[1, 1, 1], [1, 1, 1], [0, 1, 1]]),
        ),
        (
            "8192,65536",
            np.array([[0, 0, 0], [0, 2**13, 0], [0, 2**16, 2**0]]),
            np.array([[1, 1, 1], [1, 1, 1], [1, 1, 0]]),
        ),
    ],
)
def test_build_driz_weight_multiple_good_bits(
    base_image, good_bits, dq_array, expected_output
):
    data_shape = dq_array.shape
    img1 = base_image()
    img1.dq = dq_array
    img1.data = Quantity(np.ones(data_shape), unit="DN / s")

    result = resample_utils.build_driz_weight(
        img1, weight_type=None, good_bits=good_bits
    )

    np.testing.assert_array_equal(result, expected_output)


@pytest.mark.parametrize(
    "good_bits",
    [
        "~DO_NOT_USE+NON_SCIENCE",
        "~513",
        "~1+512",
        "~1,512",
        "LOW_QE+NONLINEAR",
        "73728",
        "8192+65536",
        "8192,65536",
    ],
)
def test_set_good_bits_in_resample_meta(base_image, good_bits):
    model = maker_utils.mk_level2_image(shape=(100, 100))
    model.meta.wcsinfo["vparity"] = -1

    img = datamodels.ImageModel(model)

    step = ResampleStep

    res = step.call(img, good_bits=good_bits)

    assert res.meta.resample.good_bits == good_bits


@pytest.mark.parametrize("weight_type", ["ivm", "exptime", None])
def test_build_driz_weight_different_weight_type(base_image, weight_type):
    rng = np.random.default_rng()
    img1 = base_image()
    # update attributes that will be used in building the weight array
    img1.meta.exposure.exposure_time = 10
    img1.var_rnoise = Quantity(rng.normal(1, 0.1, size=img1.shape), unit="DN2 / s2")
    # build the drizzle weight array
    result = resample_utils.build_driz_weight(
        img1, weight_type=weight_type, good_bits="~DO_NOT_USE+NON_SCIENCE"
    )

    expected_results = {
        "ivm": img1.var_rnoise.value**-1,
        "exptime": np.ones(img1.shape, dtype=img1.data.dtype)
        * img1.meta.exposure.exposure_time,
        None: np.ones(img1.shape, dtype=img1.data.dtype),
    }

    np.testing.assert_array_almost_equal(expected_results.get(weight_type), result)



"""Main entry for the association generator"""

import argparse
import logging
import os
import sys

import numpy as np

# from romancal.associations.main import  *
# from romancal.associations import (
#    __version__,
#    AssociationPool,
#    AssociationRegistry,
#    generate,
# )
from romancal.associations import __version__, config, generate
from romancal.associations.lib.dms_base import DMSAttrConstraint
from romancal.associations.lib.log_config import DMS_config, log_config
from romancal.associations.pool import AssociationPool
from romancal.associations.registry import AssociationRegistry

__all__ = ["Main"]

# Configure logging
logger = log_config(name=__package__)

# Ruleset names
DISCOVER_RULESET = "discover"
CANDIDATE_RULESET = "candidate"


class Main:
    """
    Generate Associations from an Association Pool

    Parameters
    ----------
    args : [str, ...], or None
        The command line arguments. Can be one of

        - `None`: `sys.argv` is then used.
        - `[str, ...]`: A list of strings which create the command line
          with the similar structure as `sys.argv`

    pool : None or AssociationPool
        If `None`, a pool file must be specified in the `args`.
        Otherwise, an `AssociationPool`

    Attributes
    ----------
    pool : `AssociationPool`
        The pool read in, or passed in through the parameter `pool`

    rules : `AssociationRegistry`
        The rules used for association creation.

    associations : [`Association`, ...]
        The list of generated associations.

    Notes
    -----
    Refer to the :ref:`Association Generator <associations>`
    documentation for a full description.
    """

    def __init__(self, args=None, pool=None):
        self.configure(args=args, pool=pool)

    @classmethod
    def cli(cls, args=None, pool=None):
        """Run the full association generation process

        Parameters
        ----------
        args : [str, ...], or None
            The command line arguments. Can be one of

            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`

        pool : None or AssociationPool
            If `None`, a pool file must be specified in the `args`.
            Otherwise, an `AssociationPool`

        Returns
        -------
        generator : Main
            A fully executed association generator.
        """
        generator_cli = cls(args=args, pool=pool)
        generator_cli.generate()
        generator_cli.save()
        return generator_cli

    @property
    def orphaned(self):
        """The pool of exposures that do not belong to any association."""

        not_in_asn = np.ones((len(self.pool),), dtype=bool)
        for asn in self.associations:
            try:
                indexes = [item.index for item in asn.from_items]
            except AttributeError:
                continue
            not_in_asn[indexes] = False

        orphaned = self.pool[not_in_asn]
        return orphaned

    def configure(self, args=None, pool=None):
        """Configure to prepare for generation

        Parameters
        ----------
        args : [str, ...], or None
            The command line arguments. Can be one of

            - `None`: `sys.argv` is then used.
            - `[str, ...]`: A list of strings which create the command line
              with the similar structure as `sys.argv`

        pool : None or AssociationPool
            If `None`, a pool file must be specified in the `args`.
            Otherwise, an `AssociationPool`
        """
        self.parse_args(args, has_pool=pool)
        parsed = self.parsed

        # Configure logging
        logging_config = None
        if parsed.DMS_enabled:
            logging_config = DMS_config
        logger = log_config(name=__package__, config=logging_config)
        logger.setLevel(parsed.loglevel)
        config.DEBUG = (parsed.loglevel != 0) and (parsed.loglevel <= logging.DEBUG)

        # Preamble
        logger.info("Command-line arguments: %s", parsed)
        logger.context.set("asn_candidate_ids", parsed.asn_candidate_ids)

        if pool is None:
            logger.info(f"Reading pool {parsed.pool}")
            pool = AssociationPool.read(
                parsed.pool,
                delimiter=parsed.delimiter,
                format=parsed.pool_format,
            )
        self.pool = pool

        # DMS: Add further info to logging.
        try:
            logger.context.set("program", self.pool[0]["program"])
        except KeyError:
            pass

        # Determine mode of operation. Options are
        #  1) Only specified candidates
        #  2) Only discovered associations that do not match
        #     candidate associations
        #  3) Both discovered and all candidate associations.
        logger.info("Reading rules.")

        if (
            not parsed.discover
            and not parsed.all_candidates
            and parsed.asn_candidate_ids is None
        ):
            parsed.discover = True
            parsed.all_candidates = True
        if parsed.discover or parsed.all_candidates:
            global_constraints = constrain_on_candidates(None)
        elif parsed.asn_candidate_ids is not None:
            global_constraints = constrain_on_candidates(parsed.asn_candidate_ids)

        self.rules = AssociationRegistry(
            parsed.rules,
            include_default=not parsed.ignore_default,
            global_constraints=global_constraints,
            name=CANDIDATE_RULESET,
        )
        if parsed.discover:
            self.rules.update(
                AssociationRegistry(
                    parsed.rules,
                    include_default=not parsed.ignore_default,
                    name=DISCOVER_RULESET,
                )
            )

    def generate(self):
        """Generate the associations"""
        logger.info("Generating associations.")
        parsed = self.parsed
        self.associations = generate(
            self.pool,
            self.rules,
            version_id=parsed.version_id,
            finalize=not parsed.no_finalize,
        )
        if parsed.discover:
            logger.debug(
                "# asns found before discover filtering={}".format(
                    len(self.associations)
                )
            )
            self.associations = filter_discovered_only(
                self.associations,
                DISCOVER_RULESET,
                CANDIDATE_RULESET,
                keep_candidates=parsed.all_candidates,
            )
            self.rules.Utility.resequence(self.associations)

        # Do a grand merging. This is done particularly for
        # Level2 associations.
        if parsed.merge:
            try:
                self.associations = self.rules.Utility.merge_asns(self.associations)
            except AttributeError:
                pass

        logger.debug(self.__str__())

    def parse_args(self, args=None, has_pool=False):
        """Set command line arguments

        Parameters
        ----------
        args : list, str, or None
            List of command-line arguments.
            If a string, spaces seperate the arguments.
            If None, `sys.argv` is used.

        has_pool : bool-like
            Do not require `pool` from the command line if a pool is already in hand.
        """

        if args is None:
            args = sys.argv[1:]
        if isinstance(args, str):
            args = args.split(" ")

        parser = argparse.ArgumentParser(
            description="Generate Assocation Data Products", usage="asn_generate pool"
        )
        if not has_pool:
            parser.add_argument("pool", type=str, help="Association Pool")
        op_group = parser.add_mutually_exclusive_group()
        op_group.add_argument(
            "-i",
            "--ids",
            nargs="+",
            dest="asn_candidate_ids",
            help="space-separated list of association candidate IDs to operate on.",
        )
        op_group.add_argument(
            "--discover", action="store_true", help="Produce discovered associations"
        )
        op_group.add_argument(
            "--all-candidates",
            action="store_true",
            dest="all_candidates",
            help="Produce all association candidate-specific associations",
        )
        parser.add_argument(
            "-p",
            "--path",
            type=str,
            default=".",
            help='Folder to save the associations to. Default: "%(default)s"',
        )
        parser.add_argument(
            "--save-orphans",
            dest="save_orphans",
            nargs="?",
            const="orphaned.csv",
            default=False,
            help='Save orphaned items into the specified table. Default: "%(default)s"',
        )
        parser.add_argument(
            "--version-id",
            dest="version_id",
            nargs="?",
            const=True,
            default=None,
            help=(
                "Version tag to add into association name and products."
                " If not specified, no version will be used."
                " If specified without a value, the current time is used."
                " Otherwise, the specified string will be used."
            ),
        )
        parser.add_argument(
            "-r", "--rules", action="append", help="Association Rules file."
        )
        parser.add_argument(
            "--ignore-default",
            action="store_true",
            help="Do not include default rules. -r should be used if set.",
        )
        parser.add_argument(
            "--dry-run",
            action="store_true",
            dest="dry_run",
            help="Execute but do not save results.",
        )
        parser.add_argument(
            "-d",
            "--delimiter",
            type=str,
            default="|",
            help="""Delimiter
            to use if pool files are comma-separated-value
            (csv) type files. Default: "%(default)s"
            """,
        )
        parser.add_argument(
            "--pool-format",
            type=str,
            default="ascii",
            help=(
                "Format of the pool file."
                " Any format allowed by the astropy"
                " Unified File I/O interface is allowed."
                ' Default: "%(default)s"'
            ),
        )
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_const",
            dest="loglevel",
            const=logging.INFO,
            default=logging.NOTSET,
            help="Output progress and results.",
        )
        parser.add_argument(
            "-D",
            "--debug",
            action="store_const",
            dest="loglevel",
            const=logging.DEBUG,
            help="Output detailed debugging information.",
        )
        parser.add_argument(
            "--DMS",
            action="store_true",
            dest="DMS_enabled",
            help="Running under DMS workflow conditions.",
        )
        parser.add_argument(
            "--format",
            default="json",
            help='Format of the association files. Default: "%(default)s"',
        )
        parser.add_argument(
            "--version",
            action="version",
            version=f"%(prog)s {__version__}",
            help="Version of the generator.",
        )
        parser.add_argument(
            "--no-finalize",
            action="store_true",
            help="Do not run the finalization methods on the interim associations",
        )
        parser.add_argument(
            "--merge",
            action="store_true",
            help="Merge associations into single associations with multiple products",
        )
        parser.add_argument(
            "--no-merge",
            action=DeprecateNoMerge,
            help='Deprecated: Default is to not merge. See "--merge".',
        )

        self.parsed = parser.parse_args(args=args)

    def save(self):
        """Save the associations to disk."""
        if self.parsed.dry_run:
            return

        for asn in self.associations:
            try:
                (fname, serialized) = asn.dump(format=self.parsed.format)
            except AssociationError as exception:  # noqa: F821
                logger.warning("Cannot serialize association %s", asn)
                logger.warning("Reason:", exc_info=exception)
                continue
            with open(os.path.join(self.parsed.path, fname), "w") as f:
                f.write(serialized)

        if self.parsed.save_orphans:
            self.orphaned.write(
                os.path.join(self.parsed.path, self.parsed.save_orphans),
                format="ascii",
                delimiter="|",
            )

    def __str__(self):
        result = []
        result.append(
            (
                "There where {:d} associations and {:d} orphaned items"
                " found.\nAssociations found are:"
            ).format(len(self.associations), len(self.orphaned))
        )
        for assocs in self.associations:
            result.append(assocs.__str__())

        return "\n".join(result)


def main(args=None, pool=None):
    """Command-line entrypoint for the association generator

    Wrapper around `Main.cli` so that the return is either True or an exception.

    Parameters
    ----------
    args : [str, ...], or None
        The command line arguments. Can be one of

        - `None`: `sys.argv` is then used.
        - `[str, ...]`: A list of strings which create the command line
          with the similar structure as `sys.argv`

    pool : None or AssociationPool
        If `None`, a pool file must be specified in the `args`.
        Otherwise, an `AssociationPool`
    """
    Main.cli(args, pool)


# #########
# Utilities
# #########
class DeprecateNoMerge(argparse.Action):
    """Deprecate the `--no-merge` option"""

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        super().__init__(option_strings, dest, const=True, nargs=0, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        logger.warning(
            'The "--no-merge" option is now the default and deprecated. Use "--merge"'
            " to force merging."
        )
        setattr(namespace, self.dest, values)


def constrain_on_candidates(candidates):
    """Create a constraint based on a list of candidates

    Parameters
    ----------
    candidates : (str, ...) or None
        List of candidate id's.
        If None, then all candidates are matched.
    """
    if candidates is not None and len(candidates):
        c_list = "|".join(candidates)
        values = "".join([".+(", c_list, ").+"])
    else:
        values = None
    constraint = DMSAttrConstraint(
        name="asn_candidate",
        sources=["asn_candidate"],
        value=values,
        force_unique=True,
        is_acid=True,
        evaluate=True,
    )

    return constraint


def filter_discovered_only(
    associations,
    discover_ruleset,
    candidate_ruleset,
    keep_candidates=True,
):
    """Return only those associations that have multiple candidates

    Parameters
    ----------
    associations : iterable
        The list of associations to check. The list
        is that returned by the `generate` function.

    discover_ruleset : str
        The name of the ruleset that has the discover rules

    candidate_ruleset : str
        The name of the ruleset that finds just candidates

    keep_candidates : bool
        Keep explicit candidate associations in the list.

    Returns
    -------
    iterable
        The new list of just cross candidate associations.

    Notes
    -----
    This utility is only meant to run on associations that have
    been constructed. Associations that have been Association.dump
    and then Association.load will not return proper results.
    """
    # Split the associations along discovered/not discovered lines
    asn_by_ruleset = {candidate_ruleset: [], discover_ruleset: []}
    for asn in associations:
        asn_by_ruleset[asn.registry.name].append(asn)
    candidate_list = asn_by_ruleset[candidate_ruleset]
    discover_list = asn_by_ruleset[discover_ruleset]

    # Filter out the non-unique discovered.
    for candidate in candidate_list:
        if len(discover_list) == 0:
            break
        unique_list = []
        for discover in discover_list:
            if discover != candidate:
                unique_list.append(discover)

        # Reset the discovered list to the new unique list
        # and try the next candidate.
        discover_list = unique_list

    if keep_candidates:
        discover_list.extend(candidate_list)
    return discover_list


if __name__ == "__main__":
    main()



"""Regression tests for the photom step of the Roman pipeline"""

import math

import pytest
import roman_datamodels as rdm
from astropy import units as u
from metrics_logger.decorators import metrics_logger

from romancal.step import PhotomStep
from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@metrics_logger("DMS140")
@pytest.mark.bigdata
def test_absolute_photometric_calibration(rtdata, ignore_asdf_paths):
    """DMS140 Test: Testing application of photometric correction using
    CRDS selected photom file."""

    input_data = "r0000101001001001001_01101_0001_WFI01_flat.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    # Define step (for running and log access)
    step = PhotomStep()

    #  In Wide Field Imaging mode, the DMS shall generate Level 2 science
    # data products with absolute photometry calibrated in the WFI filter
    # used for the exposure.
    step.log.info(
        "DMS140 MSG: Testing absolute photometric "
        "calibrated image data. "
        "Success is creation of a Level 2 image file with "
        "CRDS selected photom file applied."
    )

    step.log.info(f'DMS140 MSG: Image data file: {rtdata.input.rsplit("/", 1)[1]}')

    # Note: if any of the following tests fail, check for a different
    # photom match from CRDS. Values come from roman_wfi_photom_0034.asdf

    # Test PhotomStep
    output = "r0000101001001001001_01101_0001_WFI01_photom.asdf"
    rtdata.output = output
    args = ["romancal.step.PhotomStep", rtdata.input]
    step.log.info(
        "DMS140 MSG: Running photometric conversion step."
        " The first ERROR is expected, due to extra CRDS parameters"
        " not having been implemented yet."
    )
    RomanStep.from_cmdline(args)
    photom_out = rdm.open(rtdata.output)

    step.log.info(
        "DMS140 MSG: Photom step recorded as complete? :"
        f' {photom_out.meta.cal_step.photom == "COMPLETE"}'
    )
    assert photom_out.meta.cal_step.photom == "COMPLETE"

    step.log.info(
        "DMS140 MSG: Photom megajansky conversion calculated? : "
        + str(
            (photom_out.meta.photometry.conversion_megajanskys.unit == u.MJy / u.sr)
            and math.isclose(
                photom_out.meta.photometry.conversion_megajanskys.value,
                0.3324,
                abs_tol=0.0001,
            )
        )
    )
    assert photom_out.meta.photometry.conversion_megajanskys.unit == u.MJy / u.sr
    assert math.isclose(
        photom_out.meta.photometry.conversion_megajanskys.value, 0.3324, abs_tol=0.0001
    )

    step.log.info(
        "DMS140 MSG: Photom microjanskys conversion calculated? : "
        + str(
            (
                photom_out.meta.photometry.conversion_microjanskys.unit
                == u.uJy / u.arcsec**2
            )
            and (
                math.isclose(
                    photom_out.meta.photometry.conversion_microjanskys.value,
                    7.81320,
                    abs_tol=0.0001,
                )
            )
        )
    )
    assert (
        photom_out.meta.photometry.conversion_microjanskys.unit == u.uJy / u.arcsec**2
    )
    assert math.isclose(
        photom_out.meta.photometry.conversion_microjanskys.value,
        7.81320,
        abs_tol=0.0001,
    )

    step.log.info(
        "DMS140 MSG: Pixel area in steradians calculated? : "
        + str(
            (photom_out.meta.photometry.pixelarea_steradians.unit == u.sr)
            and (
                math.isclose(
                    photom_out.meta.photometry.pixelarea_steradians.value,
                    2.8083e-13,
                    abs_tol=1.0e-17,
                )
            )
        )
    )
    assert photom_out.meta.photometry.pixelarea_steradians.unit == u.sr
    assert math.isclose(
        photom_out.meta.photometry.pixelarea_steradians.value,
        2.8083e-13,
        abs_tol=1.0e-17,
    )

    step.log.info(
        "DMS140 MSG: Pixel area in square arcseconds calculated? : "
        + str(
            (photom_out.meta.photometry.pixelarea_arcsecsq.unit == u.arcsec**2)
            and (
                math.isclose(
                    photom_out.meta.photometry.pixelarea_arcsecsq.value,
                    0.011948,
                    abs_tol=1.0e-6,
                )
            )
        )
    )
    assert photom_out.meta.photometry.pixelarea_arcsecsq.unit == u.arcsec**2
    assert math.isclose(
        photom_out.meta.photometry.pixelarea_arcsecsq.value, 0.011948, abs_tol=1.0e-6
    )

    step.log.info(
        "DMS140 MSG: Photom megajansky conversion uncertainty calculated? : "
        + str(
            (
                photom_out.meta.photometry.conversion_megajanskys_uncertainty.unit
                == u.MJy / u.sr
            )
            and (
                math.isclose(
                    photom_out.meta.photometry.conversion_megajanskys_uncertainty.value,
                    0.0,
                    abs_tol=1.0e-6,
                )
            )
        )
    )
    assert (
        photom_out.meta.photometry.conversion_megajanskys_uncertainty.unit
        == u.MJy / u.sr
    )
    assert math.isclose(
        photom_out.meta.photometry.conversion_megajanskys_uncertainty.value,
        0.0,
        abs_tol=1.0e-6,
    )

    step.log.info(
        "DMS140 MSG: Photom megajansky conversion uncertainty calculated? : "
        + str(
            (
                photom_out.meta.photometry.conversion_microjanskys_uncertainty.unit
                == u.uJy / u.arcsec**2
            )
            and (
                math.isclose(
                    photom_out.meta.photometry.conversion_microjanskys_uncertainty.value,
                    0.0,
                    abs_tol=1.0e-6,
                )
            )
        )
    )
    assert (
        photom_out.meta.photometry.conversion_microjanskys_uncertainty.unit
        == u.uJy / u.arcsec**2
    )
    assert math.isclose(
        photom_out.meta.photometry.conversion_microjanskys_uncertainty.value,
        0.0,
        abs_tol=1.0e-6,
    )

    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    step.log.info(
        "DMS140 MSG: Was the proper absolute photometry calibrated image data produced?"
        f" : {diff.identical}"
    )
    assert diff.identical, diff.report()
